from __future__ import annotations

import logging
import tempfile
from enum import IntEnum
from os import fspath
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
from osgeo import gdal
from pyproj import CRS, Transformer
from shapely import to_geojson
from shapely.geometry import box

from dolphin import io
from dolphin._types import Bbox, PathOrStr

gdal.UseExceptions()

logger = logging.getLogger(__name__)


class MaskConvention(IntEnum):
    """Enum for masking conventions to indicate the nodata value as 0 or 1.

    In numpy, 1 indicates nodata
    https://numpy.org/doc/stable/reference/maskedarray.generic.html#what-is-a-masked-array

    In SNAPHU (and isce2, and some water masks), the binary mask you pass has the
    opposite convention, where 0s mean the pixel is invalid (nodata).
    """

    ZERO_IS_NODATA = 0
    ONE_IS_NODATA = 1

    # Semantic aliases for the above:
    SNAPHU = 0
    NUMPY = 1

    def get_nodata_value(self) -> int:
        """Return the nodata value for this convention."""
        return int(not self.value)


def combine_mask_files(
    mask_files: Sequence[PathOrStr],
    output_file: PathOrStr,
    dtype: str = "uint8",
    output_convention: MaskConvention = MaskConvention.ZERO_IS_NODATA,
    input_conventions: Optional[Sequence[MaskConvention]] = None,
    combine_method: str = "any",
):
    """Combine multiple mask files into a single mask file.

    Parameters
    ----------
    mask_files : list of Path or str
        list of mask files to combine.
    output_file : PathOrStr
        Path to the combined output file.
    dtype : str, optional
        Data type of output file. Default is uint8.
    output_convention : MaskConvention, optional
        Convention to use for output mask. Default is SNAPHU,
        where 0 indicates invalid pixels.
    input_conventions : list of MaskConvention, optional
        Convention to use for each input mask. Default is None,
        where it is assumed all masks use the "0 is invalid" convention.
    combine_method : str, optional, default = 'any', choices = ['any', 'all']
        Logical operation to use to combine masks. Default is 'any',
        which means the output is masked where *any* of the input masks indicated
        a masked pixel (the masked region grows larger).
        If 'all', the only pixels masked are those in which *all* input masks
        indicated a masked pixel (the masked region shrinks).

    Raises
    ------
    ValueError
        If `input_conventions` passed and is different length as `mask_files`
        If all mask_files are not the same shape

    """
    output_file = Path(output_file)
    xsize, ysize = io.get_raster_xysize(mask_files[0])

    if combine_method not in ["any", "all"]:
        msg = "combine_method must be 'any' or 'all'"
        raise ValueError(msg)

    if input_conventions is None:
        input_conventions = [MaskConvention.ZERO_IS_NODATA] * len(mask_files)
    elif isinstance(input_conventions, MaskConvention):
        input_conventions = [input_conventions] * len(mask_files)

    if len(input_conventions) != len(mask_files):
        msg = (
            f"input_conventions ({len(input_conventions)}) must have the same"
            f" length as mask_files ({len(mask_files)})"
        )
        raise ValueError(msg)

    # Uses the numpy convention (1 = invalid, 0 = valid) for combining logic
    # Loop through mask files and update the total mask
    if combine_method == "any":
        # "any" will use `logical_or` to grow the region starting empty region (as 0s)
        mask_total = np.zeros((ysize, xsize), dtype=bool)
    elif combine_method == "all":
        # "and" uses `logical_and` to shrink the full starting region (as 1s)
        mask_total = np.ones((ysize, xsize), dtype=bool)

    for input_convention, mask_file in zip(input_conventions, mask_files):
        # TODO: if we separate input missing data from mask 1/0, this changes
        mask = io.load_gdal(mask_file, masked=True).astype(bool)
        # Fill with "mask" value
        mask = mask.filled(bool(input_convention.value))
        if input_convention != MaskConvention.NUMPY:
            mask = ~mask

        if combine_method == "any":
            mask_total = np.logical_or(mask_total, mask)
        elif combine_method == "all":
            mask_total = np.logical_and(mask_total, mask)

    # Convert to output convention
    if output_convention == MaskConvention.SNAPHU:
        mask_total = ~mask_total

    io.write_arr(
        arr=mask_total.astype(dtype),
        output_name=output_file,
        like_filename=mask_files[0],
    )


def load_mask_as_numpy(mask_file: PathOrStr) -> np.ndarray:
    """Load `mask_file` and convert it to a NumPy boolean array.

    This function reads a mask file where 0 represents invalid data and 1 represents
    good data. It converts the mask to a boolean numpy array where True values
    indicate missing data (nodata) pixels, following the numpy masking convention.

    Parameters
    ----------
    mask_file : PathOrStr
        Path to the mask file. Can be a string or a Path-like object.

    Returns
    -------
    np.ndarray
        A boolean numpy array where True values indicate nodata (invalid) pixels
        and False values indicate valid data pixels.

    Notes
    -----
    The input mask file is expected to use 0 for invalid data and 1 for good data.
    The output mask follows the numpy masking convention where True indicates
    nodata and False indicates valid data.

    """
    # The mask file will by have 0s at invalid data, 1s at good
    nodata_mask = io.load_gdal(mask_file, masked=True).astype(bool).filled(False)
    # invert the mask so Trues are the missing data pixels
    nodata_mask = ~nodata_mask
    return nodata_mask


def create_bounds_mask(
    bounds: Bbox | tuple[float, float, float, float],
    output_filename: PathOrStr,
    like_filename: PathOrStr,
    bounds_epsg: int = 4326,
    overwrite: bool = False,
) -> None:
    """Create a boolean raster mask where 1 is inside the given bounds and 0 is outside.

    Parameters
    ----------
    bounds : tuple
        (min x, min y, max x, max y) of the area of interest
    like_filename : Filename
        Reference file to copy the shape, extent, and projection.
    output_filename : Filename
        Output filename for the mask
    bounds_epsg : int, optional
        EPSG code of the coordinate system of the bounds.
        Default is 4326 (lat/lon coordinates for the bounds).
    overwrite : bool, optional
        Overwrite the output file if it already exists, by default False

    """
    if Path(output_filename).exists():
        if not overwrite:
            logger.info(f"Skipping {output_filename} since it already exists.")
            return
        else:
            logger.info(f"Overwriting {output_filename} since overwrite=True.")
            Path(output_filename).unlink()

    # Transform bounds if necessary
    # Geojson default is 4326, and GDAL handles the conversion to, e.g., UTM
    if bounds_epsg != 4326:
        transformer = Transformer.from_crs(
            CRS.from_epsg(bounds_epsg), 4326, always_xy=True
        )
        bounds = transformer.transform_bounds(*bounds)

    logger.info(f"Creating mask for bounds {bounds}")

    # Create a polygon from the bounds
    bounds_poly = box(*bounds)

    # Create the output raster
    io.write_arr(
        arr=None,
        output_name=output_filename,
        dtype=bool,
        nbands=1,
        like_filename=like_filename,
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        temp_vector_file = Path(tmpdir) / "temp.geojson"
        with open(temp_vector_file, "w") as f:
            f.write(to_geojson(bounds_poly))

        # Open the input vector file
        src_ds = gdal.OpenEx(fspath(temp_vector_file), gdal.OF_VECTOR)
        dst_ds = gdal.Open(fspath(output_filename), gdal.GA_Update)

        # Now burn in the union of all polygons
        gdal.Rasterize(dst_ds, src_ds, burnValues=[1])

    logger.info(f"Created {output_filename}")



"""stitching.py: utilities for combining interferograms into larger images."""

from __future__ import annotations

import logging
import math
import subprocess
import tempfile
from datetime import datetime
from os import fspath
from pathlib import Path
from typing import Iterable, Optional, Sequence

import numpy as np
from numpy.typing import DTypeLike
from opera_utils import group_by_date
from osgeo import gdal, osr
from rasterio.warp import transform_bounds
from tqdm.contrib.concurrent import thread_map

from dolphin import io, utils
from dolphin._types import Bbox, Filename
from dolphin.io import DEFAULT_DATETIME_FORMAT

logger = logging.getLogger(__name__)


def merge_by_date(
    image_file_list: Iterable[Filename],
    file_date_fmt: str = DEFAULT_DATETIME_FORMAT,
    output_dir: Filename = ".",
    driver: str = "GTiff",
    output_suffix: str = ".int.tif",
    out_nodata: Optional[float] = 0,
    in_nodata: Optional[float] = None,
    out_bounds: Optional[Bbox] = None,
    out_bounds_epsg: Optional[int] = None,
    options: Optional[Sequence[str]] = io.DEFAULT_TIFF_OPTIONS,
    num_workers: int = 1,
    overwrite: bool = False,
) -> dict[tuple[datetime, ...], Path]:
    """Group images from the same datetime and merge into one image per datetime.

    Parameters
    ----------
    image_file_list : Iterable[Filename]
        list of paths to images.
    file_date_fmt : Optional[str]
        Format of the datetime in the filename. Default is %Y%m%d
    output_dir : Filename
        Path to output directory
    driver : str
        GDAL driver to use for output. Default is ENVI.
    output_suffix : str
        Suffix to use to output stitched filenames. Default is ".int"
    out_nodata : Optional[float | str]
        Nodata value to use for output file. Default is 0.
    in_nodata : Optional[float | str]
        Override the files' `nodata` and use `in_nodata` during merging.
    out_bounds: Optional[tuple[float]]
        if provided, forces the output image bounds to
            (left, bottom, right, top).
        Otherwise, computes from the outside of all input images.
    out_bounds_epsg: Optional[int]
        EPSG code for the `out_bounds`.
        If not provided, assumed to match the projections of `file_list`.
    options : Optional[Sequence[str]]
        Driver-specific creation options passed to GDAL.
        Default is [dolphin.io.DEFAULT_TIFF_OPTIONS][].
    num_workers : int
        Number of dates to stitch in separate threads in parallel.
        Default is 1.
    overwrite : bool
        Overwrite existing files. Default is False.

    Returns
    -------
    dict
        key: the datetime of the SLC acquisitions/datetime pair of the interferogram.
        value: the path to the stitched image

    Notes
    -----
    This function is intended to be used with filenames that contain datetime pairs
    (from interferograms).

    """
    image_path_list = [Path(f) for f in image_file_list]
    grouped_images = group_by_date(image_path_list, file_date_fmt=file_date_fmt)
    stitched_acq_times = {}
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    for dates, cur_images in grouped_images.items():
        logger.info(f"{dates}: Stitching {len(cur_images)} images.")
        if len(dates) == 2:
            date_str = utils.format_date_pair(*dates)
        elif len(dates) == 1:
            date_str = dates[0].strftime(file_date_fmt)
        else:
            msg = f"Expected 1 or 2 dates: {dates}."
            raise ValueError(msg)
        outfile = Path(output_dir) / (date_str + output_suffix)
        stitched_acq_times[dates] = outfile

    def process_date(args):
        cur_images, outfile = args
        merge_images(
            cur_images,
            outfile=outfile,
            driver=driver,
            overwrite=overwrite,
            out_nodata=out_nodata,
            out_bounds=out_bounds,
            out_bounds_epsg=out_bounds_epsg,
            in_nodata=in_nodata,
            options=options,
        )

    # loop over the merging in parallel
    thread_map(
        process_date,
        list(zip(grouped_images.values(), stitched_acq_times.values())),
        max_workers=num_workers,
        desc="Merging images by date",
    )

    return stitched_acq_times


def merge_images(
    file_list: Sequence[Filename],
    outfile: Filename,
    target_aligned_pixels: bool = True,
    out_bounds: Optional[Bbox] = None,
    out_bounds_epsg: Optional[int] = None,
    strides: Optional[dict[str, int]] = None,
    driver: str = "GTiff",
    out_nodata: Optional[float] = 0,
    out_dtype: Optional[DTypeLike] = None,
    in_nodata: Optional[float] = None,
    resample_alg: str = "lanczos",
    overwrite: bool = False,
    options: Optional[Sequence[str]] = io.DEFAULT_TIFF_OPTIONS,
    create_only: bool = False,
) -> None:
    """Combine multiple SLC images on the same date into one image.

    Parameters
    ----------
    file_list : list[Filename]
        list of raster filenames
    outfile : Filename
        Path to output file
    target_aligned_pixels: bool
        If True, adjust output image bounds so that pixel coordinates
        are integer multiples of pixel size, matching the ``-tap``
        options of GDAL utilities.
        Default is True.
    out_bounds: Optional[tuple[float]]
        if provided, forces the output image bounds to
            (left, bottom, right, top).
        Otherwise, computes from the outside of all input images.
        Note that using `resample_alg='nearest'` may result in bounds not
        equaling the exact `out_bounds` due to the nearest-neighbor
        resampling algorithm in GDAL.
    out_bounds_epsg: Optional[int]
        EPSG code for the `out_bounds`.
        If not provided, assumed to match the projections of `file_list`.
    strides : dict[str, int]
        subsample factor: {"x": x strides, "y": y strides}
    driver : str
        GDAL driver to use for output file. Default is GTiff.
    out_nodata : Optional[float | str]
        Nodata value to use for output file. Default is 0.
    out_dtype : Optional[DTypeLike]
        Output data type. Default is None, which will use the data type
        of the first image in the list.
    in_nodata : Optional[float | str]
        Override the files' `nodata` and use `in_nodata` during merging.
    resample_alg : str, default="lanczos"
        Method for gdal to use for reprojection.
        Default is lanczos (sinc-kernel)
    overwrite : bool
        Overwrite existing files. Default is False.
    options : Optional[Sequence[str]]
        Driver-specific creation options passed to GDAL.
        Default is [dolphin.io.DEFAULT_TIFF_OPTIONS][].
    create_only : bool
        If True, creates an empty output file, does not write data. Default is False.

    """
    if strides is None:
        strides = {"x": 1, "y": 1}
    if Path(outfile).exists():
        if not overwrite:
            logger.info(f"{outfile} already exists, skipping")
            return
        else:
            logger.info(f"Overwrite=True: removing {outfile}")
            Path(outfile).unlink()

    if len(file_list) == 1 and out_bounds is None:
        logger.info("Only one image, no stitching needed")
        logger.info(f"Copying {file_list[0]} to {outfile} and zeroing nodata values.")
        _copy_set_nodata(
            file_list[0],
            outfile=outfile,
            driver=driver,
            creation_options=options,
            out_nodata=out_nodata or 0,
        )
        return

    # Make sure all the files are in the same projection.
    projection = _get_mode_projection(file_list)
    # If not, warp them to the most common projection using VRT files in a tempdir
    temp_dir = tempfile.TemporaryDirectory()

    tmp_path = Path(temp_dir.name)
    if strides is not None and strides["x"] > 1 and strides["y"] > 1:
        file_list = get_downsampled_vrts(
            file_list,
            strides=strides,
            dirname=tmp_path,
        )

    warped_file_list = warp_to_projection(
        file_list,
        dirname=tmp_path,
        projection=projection,
        resample_alg=resample_alg,
    )
    # Compute output array shape. We guarantee it will cover the output
    # bounds completely
    bounds, combined_nodata = get_combined_bounds_nodata(
        *warped_file_list,
        target_aligned_pixels=target_aligned_pixels,
        out_bounds=out_bounds,
        out_bounds_epsg=out_bounds_epsg,
    )
    (xmin, ymin, xmax, ymax) = bounds
    proj_win = (xmin, ymax, xmax, ymin)  # ul_lr = ulx, uly, lrx, lry

    # Write out the files for gdal_merge using the --optfile flag
    optfile = tmp_path / "file_list.txt"
    optfile.write_text("\n".join(map(str, warped_file_list)))
    suffix = Path(outfile).suffix
    merge_output = (tmp_path / "merged").with_suffix(suffix)
    args = [
        "gdal_merge.py",
        "-quiet",
        "-o",
        merge_output,
        "--optfile",
        optfile,
        "-of",
        driver,
    ]

    if out_nodata is not None:
        args.extend(["-a_nodata", str(out_nodata)])
    if in_nodata is not None or combined_nodata is not None:
        ndv = str(in_nodata) if in_nodata is not None else str(combined_nodata)
        args.extend(["-n", ndv])
    if out_dtype is not None:
        out_gdal_dtype = gdal.GetDataTypeName(utils.numpy_to_gdal_type(out_dtype))
        args.extend(["-ot", out_gdal_dtype])
    if create_only:
        args.append("-create")
    if options is not None:
        for option in options:
            args.extend(["-co", option])

    arg_list = [str(a) for a in args]
    logger.debug(f"Running {' '.join(arg_list)}")
    subprocess.check_call(arg_list)

    # Now clip to the provided bounding box
    gdal.Translate(
        destName=fspath(outfile),
        srcDS=fspath(merge_output),
        projWin=proj_win,
        # TODO: https://github.com/OSGeo/gdal/issues/10536
        # Figure out if we really want to resample here, or just
        # do a nearest neighbor (which is default)
        resampleAlg="bilinear",
        format=driver,
        creationOptions=options,
    )

    temp_dir.cleanup()


def get_downsampled_vrts(
    filenames: Sequence[Filename],
    strides: dict[str, int],
    dirname: Filename,
) -> list[Path]:
    """Create downsampled VRTs from a list of files.

    Does not reproject, only uses `gdal_translate`.


    Parameters
    ----------
    filenames : Sequence[Filename]
        list of filenames to warp.
    strides : dict[str, int]
        subsample factor: {"x": x strides, "y": y strides}
    dirname : Filename
        The directory to write the warped files to.

    Returns
    -------
    list[Filename]
        The warped filenames.

    """
    if not filenames:
        return []
    warped_files = []
    res = _get_resolution(filenames)
    for idx, fn in enumerate(filenames):
        p = Path(fn)
        warped_fn = Path(dirname) / _get_temp_filename(p, idx, "_downsampled")
        logger.debug(f"Downsampling {p} by {strides}")
        warped_files.append(warped_fn)
        left, bottom, right, top = io.get_raster_bounds(p)
        gdal.Translate(
            fspath(warped_fn),
            fspath(p),
            format="VRT",  # Just creates a file that will warp on the fly
            resampleAlg="nearest",  # nearest neighbor for resampling
            xRes=res[0] * strides["x"],
            yRes=res[1] * strides["y"],
            projWin=(left, top, right, bottom),
        )

    return warped_files


def _get_temp_filename(fn: Path, idx: int, extra: str = ""):
    base = utils._get_path_from_gdal_str(fn).stem
    return f"{base}_{idx}{extra}.vrt"


def warp_to_projection(
    filenames: Sequence[Filename],
    dirname: Filename,
    projection: str,
    res: Optional[tuple[float, float]] = None,
    resample_alg: str = "lanczos",
) -> list[Path]:
    """Warp a list of files to `projection`.

    If the input file's projection matches `projection`, the same file is returned.
    Otherwise, a new file is created in `dirname` with the same name as the input file,
    but with '_warped' appended.

    Parameters
    ----------
    filenames : Sequence[Filename]
        list of filenames to warp.
    dirname : Filename
        The directory to write the warped files to.
    projection : str
        The desired projection, as a WKT string or 'EPSG:XXXX' string.
    res : tuple[float, float]
        The desired [x, y] resolution.
    resample_alg : str, default="lanczos"
        Method for gdal to use for reprojection.
        Default is lanczos (sinc-kernel)

    Returns
    -------
    list[Filename]
        The warped filenames.

    """
    if projection is None:
        projection = _get_mode_projection(filenames)
    if res is None:
        res = _get_resolution(filenames)

    warped_files = []
    for idx, fn in enumerate(filenames):
        p = Path(fn)
        ds = gdal.Open(fspath(p))
        proj_in = ds.GetProjection()
        if proj_in == projection:
            warped_files.append(p)
            continue
        warped_fn = Path(dirname) / _get_temp_filename(p, idx, "_warped")
        warped_fn = Path(dirname) / f"{p.stem}_{idx}_warped.vrt"
        from_srs_name = ds.GetSpatialRef().GetName()
        to_srs_name = osr.SpatialReference(projection).GetName()
        logger.info(
            f"Reprojecting {p} from {from_srs_name} to match mode projection"
            f" {to_srs_name}"
        )
        warped_files.append(warped_fn)
        gdal.Warp(
            fspath(warped_fn),
            fspath(p),
            format="VRT",  # Just creates a file that will warp on the fly
            dstSRS=projection,
            resampleAlg=resample_alg,
            targetAlignedPixels=True,  # align in multiples of dx, dy
            xRes=res[0],
            yRes=res[1],
        )

    return warped_files


def _get_mode_projection(filenames: Iterable[Filename]) -> str:
    """Get the most common projection in the list."""
    projs = [gdal.Open(fspath(fn)).GetProjection() for fn in filenames]
    return max(set(projs), key=projs.count)


def _get_resolution(filenames: Iterable[Filename]) -> tuple[float, float]:
    """Get the most common resolution in the list."""
    gts = [gdal.Open(fspath(fn)).GetGeoTransform() for fn in filenames]
    res = [(dx, dy) for (_, dx, _, _, _, dy) in gts]
    if len(set(res)) > 1:
        msg = f"The input files have different resolutions: {res}. "
        raise ValueError(msg)
    return res[0]


def get_combined_bounds_nodata(
    *filenames: Filename,
    target_aligned_pixels: bool = False,
    out_bounds: Optional[Bbox] = None,
    out_bounds_epsg: Optional[int] = None,
    strides: Optional[dict[str, int]] = None,
) -> tuple[Bbox, Optional[float]]:
    """Get the bounds and nodata of the combined image.

    Parameters
    ----------
    filenames : list[Filename]
        list of filenames to combine
    target_aligned_pixels : bool
        if True, adjust output image bounds so that pixel coordinates
        are integer multiples of pixel size, matching the `-tap` GDAL option.
    out_bounds: Optional[Bbox]
        if provided, forces the output image bounds to
            (left, bottom, right, top).
        Otherwise, computes from the outside of all input images.
    out_bounds_epsg: Optional[int]
        The EPSG of `out_bounds`. If not provided, assumed to be the same
        as the EPSG of all `*filenames`.
    strides : dict[str, int]
        subsample factor: {"x": x strides, "y": y strides}

    Returns
    -------
    bounds : Bbox
        (min_x, min_y, max_x, max_y)
    nodata : float | None
        Nodata value of the input files

    Raises
    ------
    ValueError:
        If the inputs files have different resolutions/projections/nodata values

    """
    # scan input files
    if strides is None:
        strides = {"x": 1, "y": 1}
    xs = []
    ys = []
    resolutions = set()
    projs = set()
    nodatas = set()

    # Check all files match in resolution/projection
    for fn in filenames:
        ds = gdal.Open(fspath(fn))
        left, bottom, right, top = io.get_raster_bounds(fn)
        gt = ds.GetGeoTransform()
        dx, dy = gt[1], gt[5]

        resolutions.add((abs(dx), abs(dy)))  # dy is negative for north-up
        projs.add(ds.GetProjection())

        xs.extend([left, right])
        ys.extend([bottom, top])

        nd = io.get_raster_nodata(fn)
        # Need to stringify 'nan', or it is repeatedly added
        nodatas.add(str(nd) if (nd is not None and np.isnan(nd)) else nd)

    if len(resolutions) > 1:
        msg = f"The input files have different resolutions: {resolutions}. "
        raise ValueError(msg)
    if len(projs) > 1:
        msg = f"The input files have different projections: {projs}. "
        raise ValueError(msg)
    if len(nodatas) > 1:
        msg = f"The input files have different nodata values: {nodatas}. "
        raise ValueError(msg)
    res = (abs(dx) * strides["x"], abs(dy) * strides["y"])

    if out_bounds is not None:
        if out_bounds_epsg is not None:
            dst_epsg = io.get_raster_crs(filenames[0]).to_epsg()
            bounds = Bbox(*transform_bounds(out_bounds_epsg, dst_epsg, *out_bounds))
        else:
            bounds = out_bounds
    else:
        bounds = Bbox(min(xs), min(ys), max(xs), max(ys))

    if target_aligned_pixels:
        bounds = _align_bounds(bounds, res)

    nodata = next(iter(nodatas))
    # Convert back from string "nan"
    ndv: float | None = np.nan if nodata == "nan" else nodata  # type: ignore[assignment]
    return bounds, ndv


def _align_bounds(bounds: Bbox, res: tuple[float, float]) -> Bbox:
    """Align boundary with an integer multiple of the resolution."""
    left, bottom, right, top = bounds
    left = math.floor(left / res[0]) * res[0]
    right = math.ceil(right / res[0]) * res[0]
    bottom = math.floor(bottom / res[1]) * res[1]
    top = math.ceil(top / res[1]) * res[1]
    return Bbox(left, bottom, right, top)


def _copy_set_nodata(
    infile: Filename,
    outfile: Optional[Filename] = None,
    ext: Optional[str] = None,
    in_band: int = 1,
    out_nodata: float = 0,
    driver="ENVI",
    creation_options=io.DEFAULT_ENVI_OPTIONS,
):
    """Make a copy of infile and replace NaNs/input nodata with `out_nodata`."""
    in_p = Path(infile)
    if outfile is None:
        if ext is None:
            ext = in_p.suffix
        out_dir = in_p.parent
        outfile = out_dir / (in_p.stem + "_tmp" + ext)

    ds_in = gdal.Open(fspath(infile))
    drv = gdal.GetDriverByName(driver)
    ds_out = drv.CreateCopy(fspath(outfile), ds_in, options=creation_options)

    bnd = ds_in.GetRasterBand(in_band)
    nodata = bnd.GetNoDataValue()
    arr = bnd.ReadAsArray()
    # also make sure to replace NaNs, even if nodata is not set
    mask = np.logical_or(np.isnan(arr), arr == nodata)
    arr[mask] = out_nodata

    bnd1 = ds_out.GetRasterBand(1)
    bnd1.WriteArray(arr)
    bnd1.SetNoDataValue(out_nodata)
    ds_out = bnd1 = None

    return outfile


def warp_to_match(
    input_file: Filename,
    match_file: Filename,
    output_file: Optional[Filename] = None,
    resample_alg: str = "near",
    output_format: Optional[str] = None,
) -> Path:
    """Reproject `input_file` to align with the `match_file`.

    Uses the bounds, resolution, and CRS of `match_file`.

    Parameters
    ----------
    input_file: Filename
        Path to the image to be reprojected.
    match_file: Filename
        Path to the input image to serve as a reference for the reprojected image.
        Uses the bounds, resolution, and CRS of this image.
    output_file: Filename
        Path to the output, reprojected image.
        If None, creates an in-memory warped VRT using the `/vsimem/` protocol.
    resample_alg: str, optional, default = "near"
        Resampling algorithm to be used during reprojection.
        See https://gdal.org/programs/gdalwarp.html#cmdoption-gdalwarp-r for choices.
    output_format: str, optional, default = None
        Output format to be used for the output image.
        If None, gdal will try to infer the format from the output file extension, or
        (if the extension of `output_file` matches `input_file`) use the input driver.

    Returns
    -------
    Path
        Path to the output image.
        Same as `output_file` if provided, otherwise a path to the in-memory VRT.

    """
    bounds = io.get_raster_bounds(match_file)
    crs_wkt = io.get_raster_crs(match_file).to_wkt()
    gt = io.get_raster_gt(match_file)
    resolution = (gt[1], gt[5])

    if output_file is None:
        output_file = f"/vsimem/warped_{Path(input_file).stem}.vrt"
        logger.debug(f"Creating in-memory warped VRT: {output_file}")

    if output_format is None and Path(input_file).suffix == Path(output_file).suffix:
        output_format = io.get_raster_driver(input_file)

    options = gdal.WarpOptions(
        dstSRS=crs_wkt,
        format=output_format,
        xRes=resolution[0],
        yRes=resolution[1],
        outputBounds=bounds,
        outputBoundsSRS=crs_wkt,
        resampleAlg=resample_alg,
    )
    gdal.Warp(
        fspath(output_file),
        fspath(input_file),
        options=options,
    )

    return Path(output_file)



import numpy as np
import numpy.testing as npt

from dolphin.phase_link import metrics, simulate


def _estimate_temp_coh(est, cov_matrix):
    """explicit version of one pixel's temporal coherence"""
    gamma = 0
    N = len(est)
    count = 0
    for i in range(N):
        for j in range(i + 1, N):
            theta = np.angle(cov_matrix[i, j])
            phi = np.angle(est[i] * np.conj(est[j]))

            gamma += np.exp(1j * theta) * np.exp(-1j * phi)
            count += 1
    return np.abs(gamma) / count


def test_temp_coh():
    sigmas = [0.01, 0.1, 1, 10]
    expected_tcoh_bounds = [(0.99, 1), (0.9, 1.0), (0.6, 0.9), (0.0, 0.3)]
    out_tc = []
    out_C = []
    out_truth = []
    for sigma, (t_low, t_high) in zip(sigmas, expected_tcoh_bounds):
        simulate._seed(1)
        C, truth = simulate.simulate_coh(
            num_acq=30,
            Tau0=72,
            gamma_inf=0.95,
            gamma0=0.99,
            add_signal=True,
            signal_std=sigma,
        )
        # Get the explicit for-loop version
        temp_coh = _estimate_temp_coh(np.exp(1j * truth), C)
        assert t_low <= temp_coh <= t_high
        out_tc.append(temp_coh)
        out_C.append(C)
        out_truth.append(truth)

    # Test passing one pixel's worth of data
    tc_pixel = metrics.estimate_temp_coh(np.exp(1j * truth), C)
    assert np.allclose(tc_pixel, temp_coh)

    # Test the block version of the function
    # Just repeat the first C matrix and first estimate
    C_arrays = np.array(out_C).reshape((2, 2, 30, 30))
    est_arrays = np.exp(1j * np.array(out_truth).reshape((2, 2, 30)))
    tc_image = metrics.estimate_temp_coh(est_arrays, C_arrays)

    expected_tc_image = np.array(out_tc).reshape((2, 2))
    npt.assert_array_almost_equal(tc_image, expected_tc_image)



from __future__ import annotations

import logging
from enum import IntEnum
from functools import partial
from typing import NamedTuple, Optional

import jax.numpy as jnp
import numpy as np
from jax import Array, jit, lax
from jax.scipy.linalg import cho_factor, cho_solve
from jax.typing import ArrayLike

from dolphin._types import HalfWindow, Strides
from dolphin.utils import take_looks

from . import covariance, metrics
from ._eigenvalues import eigh_largest_stack, eigh_smallest_stack
from ._ps_filling import fill_ps_pixels

logger = logging.getLogger(__name__)


DEFAULT_STRIDES = Strides(1, 1)


class PhaseLinkRuntimeError(Exception):
    """Exception raised while running the MLE solver."""


class EstimatorType(IntEnum):
    """Type of estimator used for phase linking."""

    EVD = 0
    EMI = 1


class PhaseLinkOutput(NamedTuple):
    """Output of the MLE solver."""

    cpx_phase: np.ndarray
    """Estimated linked phase."""

    temp_coh: np.ndarray
    """Temporal coherence of the optimization.
    A goodness of fit parameter from 0 to 1 at each pixel.
    """

    shp_counts: np.ndarray
    """Number of neighbor pixels used in adaptive multilooking."""

    eigenvalues: np.ndarray
    """The smallest (largest) eigenvalue resulting from EMI (EVD)."""

    estimator: np.ndarray  # dtype: np.int8
    """The estimator type used for phase linking at each pixel."""

    avg_coh: np.ndarray | None = None
    """Average coherence across dates for each SLC."""


def run_phase_linking(
    slc_stack: ArrayLike,
    half_window: HalfWindow,
    strides: Strides = DEFAULT_STRIDES,
    use_evd: bool = False,
    beta: float = 0.0,
    reference_idx: int = 0,
    nodata_mask: ArrayLike | None = None,
    ps_mask: ArrayLike | None = None,
    use_max_ps: bool = True,
    neighbor_arrays: ArrayLike | None = None,
    avg_mag: ArrayLike | None = None,
    use_slc_amp: bool = True,
    calc_average_coh: bool = False,
    baseline_lag: Optional[int] = None,
) -> PhaseLinkOutput:
    """Estimate the linked phase for a stack of SLCs.

    If passing a `ps_mask`, will combine the PS phases with the
    estimated DS phases.

    Parameters
    ----------
    slc_stack : ArrayLike
        The SLC stack, with shape (n_images, n_rows, n_cols)
    half_window : HalfWindow, or tuple[int, int]
        A (named) tuple of (y, x) sizes for the half window.
        The full window size is 2 * half_window + 1 for x, y.
    strides : tuple[int, int], optional
        The (y, x) strides (in pixels) to use for the sliding window.
        By default (1, 1)
    use_evd : bool, default = False
        Use eigenvalue decomposition on the covariance matrix instead of
        the EMI algorithm.
    beta : float, optional
        The regularization parameter, by default 0 (no regularization).
    reference_idx : int, optional
        The index of the (non compressed) reference SLC, by default 0
    nodata_mask : ArrayLike, optional
        A mask of bad/nodata pixels to ignore when estimating the covariance.
        Pixels with `True` (or 1) are ignored, by default None
        If None, all pixels are used, by default None.
    ps_mask : ArrayLike, optional
        A mask of pixels marking persistent scatterers (PS) to
        skip when multilooking.
        Pixels with `True` (or 1) are PS and will be ignored
        (combined with `nodata_mask`).
        The phase from these pixels will be inserted back
        into the final estimate directly from `slc_stack`.
    use_max_ps : bool, optional
        Whether to use the maximum PS phase for the first pixel, or average all
        PS within the look window.
        By default True.
    neighbor_arrays : ArrayLike, optional
        The neighbor arrays to use for SHP, shape = (n_rows, n_cols, *window_shape).
        If None, a rectangular window is used. By default None.
    avg_mag : ArrayLike, optional
        The average magnitude of the SLC stack, used to to find the brightest
        PS pixels to fill within each look window.
        If None, the average magnitude will be computed from `slc_stack`.
    use_slc_amp : bool, optional
        Whether to use the SLC amplitude when outputting the MLE estimate,
        or to set the SLC amplitude to 1.0. By default True.
    calc_average_coh : bool, optional, default = False
        Whether to calculate the average coherence for each SLC date.
    baseline_lag : int, optional, default=None
        lag for temporal baseline to do short temporal baseline inversion (STBAS)


    Returns
    -------
    PhaseLinkOutput:
        A Named tuple with the following fields
    linked_phase : np.ndarray[np.complex64]
        The estimated linked phase, with shape (n_images, n_rows, n_cols)
    temp_coh : np.ndarray[np.float32]
        The temporal coherence at each pixel, shape (n_rows, n_cols)
    eigenvalues : np.ndarray[np.float32]
        The smallest (largest) eigenvalue resulting from EMI (EVD).
    `avg_coh` : np.ndarray[np.float32]
        (only If `calc_average_coh` is True) the average coherence for each SLC date

    """
    _, rows, cols = slc_stack.shape
    # Common pre-processing for both CPU and GPU versions:

    # Mask nodata pixels if given
    if nodata_mask is None:
        nodata_mask = np.zeros((rows, cols), dtype=bool)
    else:
        nodata_mask = nodata_mask.astype(bool)

    # Track the PS pixels, if given, and remove them from the stack
    # This will prevent the large amplitude PS pixels from dominating
    # the covariance estimation.
    if ps_mask is None:
        ps_mask = np.zeros((rows, cols), dtype=bool)
    else:
        ps_mask = ps_mask.astype(bool)
    _raise_if_all_nan(slc_stack)

    # Make sure we also are ignoring pixels which are nans for all SLCs
    if nodata_mask.shape != (rows, cols) or ps_mask.shape != (rows, cols):
        msg = (
            f"nodata_mask.shape={nodata_mask.shape}, ps_mask.shape={ps_mask.shape},"
            f" but != SLC (rows, cols) {rows, cols}"
        )
        raise ValueError(msg)
    # for any area that has nans in the SLC stack, mark it as nodata
    nodata_mask |= np.any(np.isnan(slc_stack), axis=0)
    # Make sure the PS mask didn't have extra burst borders that are nodata here
    ps_mask[nodata_mask] = False

    # TODO: Any other masks we need?
    ignore_mask = np.logical_or.reduce((nodata_mask, ps_mask))

    # Make a copy, and set the masked pixels to np.nan
    slc_stack_masked = slc_stack.copy()
    slc_stack_masked[:, ignore_mask] = np.nan

    cpl_out = run_cpl(
        slc_stack=slc_stack_masked,
        half_window=half_window,
        strides=strides,
        use_evd=use_evd,
        beta=beta,
        reference_idx=reference_idx,
        neighbor_arrays=neighbor_arrays,
        calc_average_coh=calc_average_coh,
        baseline_lag=baseline_lag,
    )

    # Get the smaller, looked versions of the masks
    # We zero out nodata if all pixels within the window had nodata
    mask_looked = take_looks(nodata_mask, *strides, func_type="all")

    # Convert from jax array back to np
    temp_coh = np.array(cpl_out.temp_coh)

    # Set as unit-magnitude
    cpx_phase = np.exp(1j * np.angle(cpl_out.cpx_phase))
    # Fill in the PS pixels from the original SLC stack, if it was given
    if np.any(ps_mask):
        fill_ps_pixels(
            cpx_phase,
            temp_coh,
            slc_stack,
            ps_mask,
            strides,
            avg_mag,
            reference_idx,
            use_max_ps=use_max_ps,
        )

    if use_slc_amp:
        # use the amplitude from the original SLCs
        # account for the strides when grabbing original data
        # we need to match `io.compute_out_shape` here
        slcs_decimated = decimate(slc_stack, strides)
        cpx_phase = np.exp(1j * np.angle(cpx_phase)) * np.abs(slcs_decimated)

    # Finally, ensure the nodata regions are 0
    cpx_phase[:, mask_looked] = np.nan
    temp_coh[mask_looked] = np.nan

    return PhaseLinkOutput(
        cpx_phase=cpx_phase,
        temp_coh=temp_coh,
        shp_counts=np.asarray(cpl_out.shp_counts),
        # Convert the rest to numpy for writing
        eigenvalues=np.asarray(cpl_out.eigenvalues),
        estimator=np.asarray(cpl_out.estimator),
        avg_coh=cpl_out.avg_coh,
    )


def run_cpl(
    slc_stack: np.ndarray,
    half_window: HalfWindow,
    strides: Strides,
    use_evd: bool = False,
    beta: float = 0,
    reference_idx: int = 0,
    neighbor_arrays: Optional[np.ndarray] = None,
    calc_average_coh: bool = False,
    baseline_lag: Optional[int] = None,
) -> PhaseLinkOutput:
    """Run the Combined Phase Linking (CPL) algorithm.

    Estimates a coherence matrix for each SLC pixel, then
    runs the EMI/EVD solver.

    Parameters
    ----------
    slc_stack : np.ndarray
        The SLC stack, with shape (n_slc, n_rows, n_cols)
    half_window : HalfWindow, or tuple[int, int]
        A (named) tuple of (y, x) sizes for the half window.
        The full window size is 2 * half_window + 1 for x, y.
    strides : tuple[int, int], optional
        The (y, x) strides (in pixels) to use for the sliding window.
        By default (1, 1)
    use_evd : bool, default = False
        Use eigenvalue decomposition on the covariance matrix instead of
        the EMI algorithm.
    beta : float, optional
        The regularization parameter, by default 0 (no regularization).
    reference_idx : int, optional
        The index of the (non compressed) reference SLC, by default 0
    use_slc_amp : bool, optional
        Whether to use the SLC amplitude when outputting the MLE estimate,
        or to set the SLC amplitude to 1.0. By default True.
    neighbor_arrays : np.ndarray, optional
        The neighbor arrays to use for SHP, shape = (n_rows, n_cols, *window_shape).
        If None, a rectangular window is used. By default None.
    calc_average_coh : bool, default=False
        If requested, the average of each row of the covariance matrix is computed
        for the purposes of finding the best reference (highest coherence) date
    baseline_lag : int, optional, default=None
        StBAS parameter to include only nearest-N interferograms for phase linking.
        A `baseline_lag` of `n` will only include the closest `n` interferograms.
        `baseline_line` must be positive.

    Returns
    -------
    cpx_phase : Array
        Optimized SLC phase, shape same as `slc_stack` unless Strides are requested.
    temp_coh : Array
        Temporal coherence of the optimization.
        A goodness of fit parameter from 0 to 1 at each pixel.
        shape = (out_rows, out_cols)
    eigenvalues : Array
        The eigenvalues of the coherence matrices.
        If `use_evd` is True, these are the largest eigenvalues;
        Otherwise, for EMI they are the smallest.
        shape = (out_rows, out_cols)
    estimator : Array
        The estimator used at each pixel.
        0 = EVD, 1 = EMI
        shape = (out_rows, out_cols)
    avg_coh : np.ndarray | None
        The average coherence of each row of the coherence matrix,
        if requested.
        shape = (nslc, out_rows, out_cols)

    """
    C_arrays = covariance.estimate_stack_covariance(
        slc_stack,
        half_window,
        strides,
        neighbor_arrays=neighbor_arrays,
    )
    ns = slc_stack.shape[0]
    if baseline_lag:
        u_rows, u_cols = jnp.triu_indices(ns, baseline_lag)
        l_rows, l_cols = jnp.tril_indices(ns, -baseline_lag)
        C_arrays = C_arrays.at[:, :, u_rows, u_cols].set(0.0 + 0j)
        C_arrays = C_arrays.at[:, :, l_rows, l_cols].set(0.0 + 0j)

    cpx_phase, eigenvalues, estimator = process_coherence_matrices(
        C_arrays,
        use_evd=use_evd,
        beta=beta,
        reference_idx=reference_idx,
    )
    # Get the temporal coherence
    temp_coh = metrics.estimate_temp_coh(cpx_phase, C_arrays)

    # Reshape the (rows, cols, nslcs) output to be same as input stack
    cpx_phase_reshaped = jnp.moveaxis(cpx_phase, -1, 0)

    # Get the SHP counts for each pixel (if not using Rect window)
    if neighbor_arrays is None:
        shp_counts = jnp.zeros(temp_coh.shape, dtype=np.int16)
    else:
        shp_counts = jnp.sum(neighbor_arrays, axis=(-2, -1))

    if calc_average_coh:
        # If requested, average the Cov matrix at each row for reference selection
        avg_coh_per_date = jnp.abs(C_arrays).mean(axis=3)
        avg_coh = np.argmax(avg_coh_per_date, axis=2)
    else:
        avg_coh = None

    return PhaseLinkOutput(
        cpx_phase=cpx_phase_reshaped,
        temp_coh=temp_coh,
        shp_counts=shp_counts,
        eigenvalues=eigenvalues,
        estimator=estimator,
        avg_coh=avg_coh,
    )


@partial(jit, static_argnames=("use_evd", "beta", "reference_idx"))
def process_coherence_matrices(
    C_arrays,
    use_evd: bool = False,
    beta: float = 0.0,
    reference_idx: int = 0,
) -> tuple[Array, Array, Array]:
    """Estimate the linked phase for a stack of coherence matrices.

    This function is used after coherence estimation to estimate the
    optimized SLC phase.

    Parameters
    ----------
    C_arrays : ndarray, shape = (rows, cols, nslc, nslc)
        The sample coherence matrix at each pixel
        (e.g. from [dolphin.phase_link.covariance.estimate_stack_covariance][])
    use_evd : bool, default = False
        Use eigenvalue decomposition on the covariance matrix instead of
        the EMI algorithm of [@Ansari2018EfficientPhaseEstimation].
    beta : float, optional
        The regularization parameter for inverting Gamma = |C|
        The regularization is applied as (1 - beta) * Gamma + beta * I
        Default is 0 (no regularization).
    reference_idx : int, optional
        The index of the reference acquisition, by default 0
        All outputs are multiplied by the conjugate of the data at this index.

    Returns
    -------
    eig_vecs : ndarray[float32], shape = (rows, cols, nslc)
        The phase resulting from the optimization at each output pixel.
        Shape is same as input slcs unless Strides > (1, 1)
    eig_vals : ndarray[float], shape = (rows, cols)
        The smallest (largest) eigenvalue as solved by EMI (EVD).
    estimator : Array
        The estimator used at each pixel.
        0 = EVD, 1 = EMI

    """
    rows, cols, n, _ = C_arrays.shape

    if use_evd:
        # EVD
        evd_eig_vals, evd_eig_vecs = eigh_largest_stack(C_arrays)
        eig_vals, eig_vecs = evd_eig_vals, evd_eig_vecs
        estimator = jnp.zeros(eig_vals.shape, dtype=bool)
    else:
        # EMI
        # estimate the wrapped phase based on the EMI paper
        # *smallest* eigenvalue decomposition of the (|Gamma|^-1  *  C) matrix

        # Identity used for regularization and for solving
        Id = jnp.eye(n, dtype=C_arrays.dtype)
        # repeat the identity matrix for each pixel
        Id = jnp.tile(Id, (rows, cols, 1, 1))

        Gamma = jnp.abs(C_arrays)
        if beta > 0:
            # Perform regularization
            Gamma = (1 - beta) * Gamma + beta * Id

        # Attempt to invert Gamma
        cho, is_lower = cho_factor(Gamma)

        # Check: If it fails the cholesky factor, it's close to singular and
        # we should just fall back to EVD
        # Use the already- factored |Gamma|^-1, solving Ax = I gives the inverse
        Gamma_inv = cho_solve((cho, is_lower), Id)
        # We're looking for the lambda nearest to 1. So shift by 0.99
        # Also, use the evd vectors as iteration starting point:
        mu = 0.99
        emi_eig_vals, emi_eig_vecs = eigh_smallest_stack(Gamma_inv * C_arrays, mu)
        # From the EMI paper, normalize the eigenvectors to have norm sqrt(n)
        emi_eig_vecs = (
            jnp.sqrt(n)
            * emi_eig_vecs
            / jnp.linalg.norm(emi_eig_vecs, axis=-1, keepdims=True)
        )
        # is the output is the inverse of the eigenvectors? or inverse conj?

        # Use https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.select.html
        # Note that `if` would fail the jit tracing
        # https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#cond
        inv_has_nans = jnp.any(jnp.isnan(Gamma_inv), axis=(-1, -2))

        # Must broadcast the 2D boolean array so it's the same size as the outputs
        inv_has_nans_3d = jnp.tile(inv_has_nans[:, :, None], (1, 1, n))

        # For EVD, or places where inverting |Gamma| failed: fall back to computing EVD
        evd_eig_vals, evd_eig_vecs = eigh_largest_stack(C_arrays)
        eig_vecs = lax.select(
            inv_has_nans_3d,
            # Run this on True: EVD, since we failed to invert:
            evd_eig_vecs,
            # Otherwise, on False, we're fine to use EMI
            emi_eig_vecs,
        )

        eig_vals = lax.select(inv_has_nans, evd_eig_vals, emi_eig_vals)
        # Make array of ints to indicate which estimator was used for each pixel
        # 0 means EVD, 1 mean EMI
        evd_used = jnp.zeros(emi_eig_vals.shape, dtype=jnp.int8)
        emi_used = jnp.ones(emi_eig_vals.shape, dtype=jnp.int8)
        estimator = lax.select(inv_has_nans, evd_used, emi_used)

    # Now the shape of eig_vecs is (rows, cols, nslc)
    # at pixel (r, c), eig_vecs[r, c] is the largest (smallest) eigenvector if
    # we picked EVD (EMI)
    # The phase estimate on the reference day will be size (rows, cols)
    ref = eig_vecs[:, :, reference_idx]
    # Make sure each still has 3 dims, then reference all phases to `ref`
    evd_estimate = eig_vecs * jnp.exp(-1j * jnp.angle(ref[:, :, None]))

    return evd_estimate, eig_vals, estimator.astype("uint8")


def decimate(arr: ArrayLike, strides: Strides) -> Array:
    """Decimate an array by strides in the x and y directions.

    Output will match [`io.compute_out_shape`][dolphin.io.compute_out_shape]

    Parameters
    ----------
    arr : ArrayLike
        2D or 3D array to decimate.
    strides : dict[str, int]
        The strides in the x and y directions.

    Returns
    -------
    ArrayLike
        The decimated array.

    """
    ys, xs = strides
    rows, cols = arr.shape[-2:]
    start_r = ys // 2
    start_c = xs // 2
    end_r = (rows // ys) * ys + 1
    end_c = (cols // xs) * xs + 1
    return arr[..., start_r:end_r:ys, start_c:end_c:xs]


def _raise_if_all_nan(slc_stack: np.ndarray):
    """Check for all NaNs in each SLC of the stack."""
    nans = np.isnan(slc_stack)
    # Check that there are no SLCS which are all nans:
    bad_slc_idxs = np.where(np.all(nans, axis=(1, 2)))[0]
    if bad_slc_idxs.size > 0:
        msg = f"slc_stack[{bad_slc_idxs}] out of {len(slc_stack)} are all NaNs."
        raise PhaseLinkRuntimeError(msg)



from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from dolphin import Strides, io
from dolphin.io import VRTStack
from dolphin.utils import compute_out_shape


class TestLoad:
    def test_load(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        assert arr.shape == (100, 200)

    def test_get_raster_xysize(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        assert arr.shape == (100, 200)
        assert io.get_raster_xysize(raster_100_by_200) == (200, 100)

    def test_load_1_slice(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        block = io.load_gdal(raster_100_by_200, rows=slice(0, 10))
        assert block.shape == (10, 200)
        npt.assert_allclose(block, arr[:10, :])

        block = io.load_gdal(raster_100_by_200, cols=slice(10, 20))
        assert block.shape == (100, 10)
        npt.assert_allclose(block, arr[:, 10:20])

    def test_load_slices(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        block = io.load_gdal(raster_100_by_200, rows=slice(0, 10), cols=slice(0, 10))
        assert block.shape == (10, 10)
        npt.assert_allclose(block, arr[:10, :10])

        block = io.load_gdal(raster_100_by_200, rows=slice(10, 20), cols=slice(10, 20))
        assert block.shape == (10, 10)
        npt.assert_allclose(block, arr[10:20, 10:20])

    def test_load_none_slices(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        block = io.load_gdal(raster_100_by_200, rows=slice(0, 10), cols=slice(None))
        assert block.shape == (10, 200)
        npt.assert_allclose(block, arr[:10, :])

        block = io.load_gdal(
            raster_100_by_200, rows=slice(None, None, None), cols=slice(10, 20)
        )
        assert block.shape == (100, 10)
        npt.assert_allclose(block, arr[:, 10:20])

    def test_load_slice_oob(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200)
        block = io.load_gdal(raster_100_by_200, rows=slice(0, 300), cols=slice(0, 300))
        assert block.shape == (100, 200)
        npt.assert_allclose(block, arr)

        with pytest.raises(IndexError):
            block = io.load_gdal(
                raster_100_by_200, rows=slice(300, 400), cols=slice(0, 10)
            )

    def test_load_masked(self, raster_with_nan_block):
        arr = io.load_gdal(raster_with_nan_block, masked=True)
        assert isinstance(arr, np.ma.masked_array)
        assert np.ma.is_masked(arr)
        assert arr[arr.mask].size == 32 * 32
        assert np.all(arr.mask[:32, :32])

        arr = io.load_gdal(raster_with_nan_block)
        assert not isinstance(arr, np.ma.masked_array)
        assert not np.ma.is_masked(arr)
        assert np.all(np.isnan(arr[:32, :32]))

    def test_load_masked_empty_nodata(self, raster_100_by_200):
        arr = io.load_gdal(raster_100_by_200, masked=True)
        assert isinstance(arr, np.ma.masked_array)
        assert arr.mask == np.ma.nomask

    def test_load_band(self, tmp_path, slc_stack, slc_file_list):
        # Check on a VRT, which has multiple bands
        vrt_file = tmp_path / "test.vrt"
        s = VRTStack(slc_file_list, outfile=vrt_file)

        assert s.shape == slc_stack.shape
        assert len(s) == len(slc_stack) == len(slc_file_list)

        arr = io.load_gdal(s.outfile)
        npt.assert_array_equal(arr, slc_stack)

        # Now load each band
        for i in range(len(slc_stack)):
            layer = io.load_gdal(s.outfile, band=i + 1)
            npt.assert_array_equal(layer, slc_stack[i])


def test_get_raster_bounds(slc_file_list_nc_wgs84):
    # Use the created WGS84 SLC
    bnds = io.get_raster_bounds(slc_file_list_nc_wgs84[0])
    expected = (-5.5, -2.0, 4.5, 3.0)
    assert bnds == expected


class TestWriteArr:
    def test_write_arr_like(self, raster_100_by_200, tmpdir):
        arr = io.load_gdal(raster_100_by_200)

        ones = np.ones_like(arr)
        save_name = tmpdir / "ones.tif"
        io.write_arr(arr=ones, like_filename=raster_100_by_200, output_name=save_name)

        ones_loaded = io.load_gdal(save_name)
        npt.assert_array_almost_equal(ones, ones_loaded)

    def test_write_empty_like(self, raster_100_by_200, tmpdir):
        save_name = tmpdir / "empty.tif"
        io.write_arr(arr=None, like_filename=raster_100_by_200, output_name=save_name)

        empty_loaded = io.load_gdal(save_name)
        zeros = np.zeros_like(empty_loaded)
        npt.assert_array_almost_equal(empty_loaded, zeros)

    def test_write_metadata(self, raster_100_by_200, tmpdir):
        save_name = tmpdir / "empty_nometa.tif"
        io.write_arr(arr=None, like_filename=raster_100_by_200, output_name=save_name)
        assert io.get_raster_dtype(save_name) == np.complex64
        assert io.get_raster_nodata(save_name) is None

        save_name = tmpdir / "empty_bool_255_nodata.tif"
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            output_name=save_name,
            dtype=bool,
            nodata=255,
        )
        assert io.get_raster_nodata(save_name) == 255

        save_name = tmpdir / "empty_nan_nodata.tif"
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            output_name=save_name,
            nodata=np.nan,
        )
        assert np.isnan(io.get_raster_nodata(save_name))

    def test_write_units(self, raster_100_by_200, tmpdir):
        save_name = tmpdir / "empty_nometa.tif"
        io.write_arr(arr=None, like_filename=raster_100_by_200, output_name=save_name)
        assert io.get_raster_units(save_name) is None

        save_name = tmpdir / "empty_nan_meters.tif"
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            output_name=save_name,
            units="meters",
        )
        assert io.get_raster_units(save_name) == "meters"

    def test_save_strided_1(self, raster_100_by_200, tmpdir):
        save_name = tmpdir / "same_size.tif"
        strides = Strides(1, 1)
        out_shape = compute_out_shape((100, 200), strides)
        assert out_shape == (100, 200)
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            shape=out_shape,
            output_name=save_name,
        )

        xsize, ysize = io.get_raster_xysize(save_name)
        assert (ysize, xsize) == (100, 200)

    def test_save_strided_2(self, raster_100_by_200, tmpdir):
        save_name2 = tmpdir / "smaller_size.tif"
        strides = Strides(4, 2)
        out_shape = compute_out_shape((100, 200), strides)
        assert out_shape == (25, 100)
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            shape=out_shape,
            output_name=save_name2,
        )
        xsize, ysize = io.get_raster_xysize(save_name2)
        assert (ysize, xsize) == (25, 100)

    def test_save_block(self, raster_100_by_200, tmpdir):
        save_name = tmpdir / "empty.tif"
        io.write_arr(arr=None, like_filename=raster_100_by_200, output_name=save_name)

        block_loaded = io.load_gdal(save_name)
        arr = np.zeros_like(block_loaded)
        npt.assert_array_almost_equal(block_loaded, arr)

        io.write_block(
            cur_block=np.ones((20, 30)),
            filename=save_name,
            row_start=0,
            col_start=0,
        )
        block_loaded2 = io.load_gdal(save_name)
        arr[:20, :30] = 1
        npt.assert_array_almost_equal(block_loaded2, arr)

        io.write_block(
            cur_block=np.ones((20, 30)),
            filename=save_name,
            row_start=0,
            col_start=0,
        )
        block_loaded2 = io.load_gdal(save_name)
        arr[:20, :30] = 1
        npt.assert_array_almost_equal(block_loaded2, arr)

    @pytest.fixture()
    def cpx_arr(self, shape=(100, 200)):
        rng = np.random.default_rng()
        arr = rng.normal(size=shape) + 1j * rng.normal(size=shape)
        return arr.astype(np.complex64)

    def test_save_cpx(self, raster_100_by_200, cpx_arr, tmpdir):
        save_name = tmpdir / "complex.tif"
        io.write_arr(
            arr=cpx_arr, like_filename=raster_100_by_200, output_name=save_name
        )
        arr_loaded = io.load_gdal(save_name)
        assert arr_loaded.dtype == np.complex64
        npt.assert_array_almost_equal(arr_loaded, cpx_arr)

    def test_save_block_cpx(self, raster_100_by_200, cpx_arr, tmpdir):
        save_name = tmpdir / "complex_block.tif"
        # Start with empty file
        io.write_arr(
            arr=None,
            like_filename=raster_100_by_200,
            output_name=save_name,
            dtype=np.complex64,
        )
        arr_loaded = io.load_gdal(save_name)
        assert (arr_loaded == 0).all()

        io.write_block(
            cur_block=np.ones((20, 30), dtype=np.complex64),
            filename=save_name,
            row_start=0,
            col_start=0,
        )
        arr_loaded = io.load_gdal(save_name)
        assert (arr_loaded[:20, :30] == 1 + 0j).all()
        assert (arr_loaded[20:, 30:] == 0).all()

        block_cpx = cpx_arr[:10, :10].copy()
        io.write_block(
            cur_block=block_cpx,
            filename=save_name,
            row_start=20,
            col_start=20,
        )
        arr_loaded = io.load_gdal(save_name)
        assert (arr_loaded[20:30, 20:30] == block_cpx).all()


def test_get_raster_block_sizes(raster_100_by_200, tiled_raster_100_by_200):
    assert io.get_raster_chunk_size(tiled_raster_100_by_200) == [32, 32]
    assert io.get_raster_chunk_size(raster_100_by_200) == [200, 1]


def test_format_nc_filename():
    expected = 'NETCDF:"/usr/19990101/20200303_20210101.nc":"//variable"'
    assert (
        io.format_nc_filename("/usr/19990101/20200303_20210101.nc", "variable")
        == expected
    )

    # check on Path
    assert (
        io.format_nc_filename(Path("/usr/19990101/20200303_20210101.nc"), "variable")
        == expected
    )

    # check non-netcdf file
    assert (
        io.format_nc_filename("/usr/19990101/20200303_20210101.tif")
        == "/usr/19990101/20200303_20210101.tif"
    )
    assert (
        io.format_nc_filename("/usr/19990101/20200303_20210101.int", "ignored")
        == "/usr/19990101/20200303_20210101.int"
    )

    with pytest.raises(ValueError):
        # Missing the subdataset name
        io.format_nc_filename("/usr/19990101/20200303_20210101.nc")



from __future__ import annotations

import datetime
import logging
from abc import ABC, abstractmethod
from pathlib import Path

import numpy as np
import xarray
from pyproj import CRS, Transformer
from shapely.geometry import box

import dolphin.atmosphere._utils as utils
import dolphin.atmosphere.model_levels as ml

logger = logging.getLogger(__name__)

_ZMIN = np.float64(-100)  # minimum required height
_ZREF = np.float64(26000)  # maximum integration height
# when not specified by user
_g0 = g0 = np.float64(9.80665)  # Standard gravitational constant
G1 = np.float64(9.80616)  # Gravitational constant @ 45° latitude used for corrections
# of earth's centrifugal force

TIME_RES = {
    "ECMWF": 1,
    "HRES": 6,
}


class WeatherModel(ABC):
    """Implement a generic weather model for getting estimated SAR delays."""

    def __init__(self):
        """Initiate the weather model."""
        # Initialize model-specific constants/parameters
        self._k1 = None
        self._k2 = None
        self._k3 = None
        self._humidityType = "q"
        self._a = []
        self._b = []

        self.files = None

        self._time_res = None  # time resolution of the weather model in hours

        self._lon_res = None
        self._lat_res = None
        self._x_res = None
        self._y_res = None

        self._classname = None
        self._dataset = None
        self._Name = None
        self._wmLoc = None

        self._model_level_type = "ml"

        self._valid_range = (
            datetime.datetime(1900, 1, 1).replace(
                tzinfo=datetime.timezone(offset=datetime.timedelta())
            ),
            datetime.datetime.now(datetime.timezone.utc).date(),
        )  # Tuple of min/max years where data is available.
        self._lag_time = datetime.timedelta(days=30)  # Availability lag time in days
        self._time = None
        self._bbox = None

        # Define fixed constants
        self._R_v = 461.524
        self._R_d = 287.06  # in our original code this was 287.053
        self._g0 = _g0  # gravity constant
        self._zmin = _ZMIN  # minimum integration height
        self._zmax = _ZREF  # max integration height
        self._proj = None

        # setup data structures
        self._levels = None
        self._xs = np.empty((1, 1, 1))  # Use generic x/y/z instead of lon/lat/height
        self._ys = np.empty((1, 1, 1))
        self._zs = np.empty((1, 1, 1))

        self._lats = None
        self._lons = None
        self._ll_bounds = None
        self._valid_bounds = box(-180, -90, 180, 90)  # Shapely box with WSEN bounds

        self._p = None
        self._q = None
        self._rh = None
        self._t = None
        self._e = None
        self._wet_refractivity = None
        self._hydrostatic_refractivity = None
        self._wet_ztd = None
        self._hydrostatic_ztd = None

    def __str__(self):
        string = "\n"
        string += "======Weather Model class object=====\n"
        string += "Weather model time: {}\n".format(self._time)
        string += "Latitude resolution: {}\n".format(self._lat_res)
        string += "Longitude resolution: {}\n".format(self._lon_res)
        string += "Native projection: {}\n".format(self._proj)
        string += "ZMIN: {}\n".format(self._zmin)
        string += "ZMAX: {}\n".format(self._zmax)
        string += "k1 = {}\n".format(self._k1)
        string += "k2 = {}\n".format(self._k2)
        string += "k3 = {}\n".format(self._k3)
        string += "Humidity type = {}\n".format(self._humidityType)
        string += "=====================================\n"
        string += "Class name: {}\n".format(self._classname)
        string += "Dataset: {}\n".format(self._dataset)
        string += "=====================================\n"
        string += "A: {}\n".format(self._a)
        string += "B: {}\n".format(self._b)
        if self._p is not None:
            string += "Number of points in Lon/Lat = {}/{}\n".format(*self._p.shape[:2])
            string += "Total number of grid points (3D): {}\n".format(
                np.prod(self._p.shape)
            )
        if self._xs.size == 0:
            string += "Minimum/Maximum y: {: 4.2f}/{: 4.2f}\n".format(
                np.nanmin(self._ys), np.nanmax(self._ys)
            )
            string += "Minimum/Maximum x: {: 4.2f}/{: 4.2f}\n".format(
                np.nanmin(self._xs), np.nanmax(self._xs)
            )
            string += "Minimum/Maximum zs/heights: {: 10.2f}/{: 10.2f}\n".format(
                np.nanmin(self._zs), np.nanmax(self._zs)
            )
        string += "=====================================\n"
        return str(string)

    def _get_time(self):
        return self._time

    def set_time(self, time, fmt="%Y-%m-%dT%H:%M:%S"):
        """Set the time for a weather model."""
        if isinstance(time, str):
            self._time = datetime.datetime.strptime(time, fmt)
        elif isinstance(time, datetime.datetime):
            self._time = time
        else:
            raise TypeError('"time" must be a string or a datetime object')
        if self._time.tzinfo is None:
            self._time = self._time.replace(
                tzinfo=datetime.timezone(offset=datetime.timedelta())
            )

    def get_latlon_bounds(self):
        """Get the lat lon bounding box."""
        return self._ll_bounds

    def set_latlon_bounds(self, ll_bounds, Nextra=2, output_spacing=None):
        """Correct lat/lon bounds.

        because not all of the weather models have valid data exactly
        bounded by -90/90 (lats) and -180/180 (lons); for GMAO and MERRA2,
        need to adjust the longitude higher end with an extra buffer;
        for other models, the exact bounds are close to  -90/90 (lats)
        and -180/180 (lons) and thus can be rounded to the above regions
        (either in the downloading-file API or subsetting-data API)
        without problems.
        """
        ex_buffer_lon_max = 0.0

        if self._Name in "HRRR HRRR-AK HRES".split():
            Nextra = 6  # have a bigger buffer

        else:
            ex_buffer_lon_max = self._lon_res

        # At boundary lats and lons, need to modify Nextra buffer
        # so that the lats and lons do not exceed the boundary
        S, N, W, E = ll_bounds

        # Adjust bounds if they get near the poles or IDL
        pixlat, pixlon = Nextra * self._lat_res, Nextra * self._lon_res

        S = np.max([S - pixlat, -90.0 + pixlat])
        N = np.min([N + pixlat, 90.0 - pixlat])
        W = np.max(
            [W - (pixlon + ex_buffer_lon_max), -180.0 + (pixlon + ex_buffer_lon_max)]
        )
        E = np.min(
            [E + (pixlon + ex_buffer_lon_max), 180.0 - pixlon - ex_buffer_lon_max]
        )
        if output_spacing is not None:
            S, N, W, E = [
                np.floor(S / output_spacing) * output_spacing,
                np.ceil(N / output_spacing) * output_spacing,
                np.floor(W / output_spacing) * output_spacing,
                np.ceil(E / output_spacing) * output_spacing,
            ]

        self._ll_bounds = np.array([S, N, W, E])

    def get_wm_loc(self):
        """Get the path to the direct with the weather model files."""
        wmLoc = Path.cwd() / "weather_files" if self._wmLoc is None else self._wmLoc
        return wmLoc

    def set_wm_loc(self, weather_model_directory: str):
        """Set the path to the directory with the weather model files."""
        self._wmLoc = weather_model_directory

    def load(self, _zlevels=None):
        """Call the load_weather method.

        Each model class should define a load_weather.
        method appropriate for that class.
        """
        # If the weather file has already been processed, do nothing
        outLoc = self.get_wm_loc()
        path_wm_raw = Path(outLoc) / "{}_{}.{}".format(
            self._Name,
            datetime.datetime.strftime(self._get_time(), "%Y_%m_%d_T%H_%M_%S"),
            "nc",
        )

        self._out_name = self.out_file(outLoc)

        if Path(self._out_name).exists():
            return self._out_name
        else:
            # Load the weather just for the query points
            self.load_weather(f=path_wm_raw)

            # Process the weather model data
            self._find_e()
            self._uniform_in_z(_zlevels=_zlevels)
            self._check_for_nans()

            # Compute Zenith delays at the weather model grid nodes
            self._get_ztd()
            return None

    @abstractmethod
    def load_weather(self, f=None):
        """Load weather model.

        Should be implemented in each weather model type class.
        """

    def set_level_type(self, levelType):
        """Set the level type to model levels or pressure levels."""
        if levelType in "ml pl nat prs".split():
            self._model_level_type = levelType
        else:
            raise RuntimeError(f"Level type {levelType} is not recognized")

        if levelType in "ml nat".split():
            self.__model_levels__()
        else:
            self.__pressure_levels__()

    # def _convertmb2Pa(self, pres):
    #     """Convert pressure in millibars to Pascals."""
    #     return 100 * pres

    def _get_heights(self, lats, geo_hgt, geo_ht_fill=np.nan):
        """Transform geo heights to WGS84 ellipsoidal heights."""
        geo_ht_fix = np.where(geo_hgt != geo_ht_fill, geo_hgt, np.nan)
        lats_full = np.broadcast_to(lats[..., np.newaxis], geo_ht_fix.shape)
        self._zs = utils.geo_to_ht(lats_full, geo_ht_fix)

    def _find_e(self):
        """Check the type of e-calculation needed."""
        if self._humidityType == "rh":
            self._find_e_from_rh()
        elif self._humidityType == "q":
            self._find_e_from_q()
        else:
            raise RuntimeError("Not a valid humidity type")
        self._rh = None
        self._q = None

    def _find_e_from_q(self):
        """Calculate e, partial pressure of water vapor."""
        svp = utils.find_svp(self._t)
        # We have q = w/(w + 1), so w = q/(1 - q)
        w = self._q / (1 - self._q)
        self._e = w * self._R_v * (self._p - svp) / self._R_d

    def _find_e_from_rh(self):
        """Calculate partial pressure of water vapor."""
        svp = utils.find_svp(self._t)
        self._e = self._rh / 100 * svp

    def get_wet_refractivity(self):
        """Calculate the wet delay from pressure, temperature, and e."""
        self._wet_refractivity = (
            self._k2 * self._e / self._t + self._k3 * self._e / self._t**2
        )

    def get_hydro_refractivity(self):
        """Calculate the hydrostatic delay from pressure and temperature."""
        self._hydrostatic_refractivity = self._k1 * self._p / self._t

    def _get_wet_refractivity(self):
        return self._wet_refractivity

    def _get_hydro_refractivity(self):
        return self._hydrostatic_refractivity

    def _adjust_grid(self, ll_bounds=None):
        """Pad the weather grid with a level at self._zmin.

        if it does not already go that low.
        <<The functionality below has been removed.>>
        <<It also removes levels that are above self._zmax,
        since they are not needed.>>.
        """
        if self._zmin < np.nanmin(self._zs):
            # first add in a new layer at zmin
            self._zs = np.insert(self._zs, 0, self._zmin)

            self._p = utils.pad_lower(self._p)
            self._t = utils.pad_lower(self._t)
            self._e = utils.pad_lower(self._e)
            self._wet_refractivity = utils.pad_lower(self._wet_refractivity)
            self._hydrostatic_refractivity = utils.pad_lower(
                self._hydrostatic_refractivity
            )
            if ll_bounds is not None:
                self._trim_extent(ll_bounds)

    def _get_ztd(self):
        """Compute the full slant tropospheric delay for each weather model grid node.

        using the reference height zref.
        """
        self.get_wet_refractivity()
        self.get_hydro_refractivity()
        self._adjust_grid(self.get_latlon_bounds())

        wet = self._get_wet_refractivity()
        hydro = self._get_hydro_refractivity()

        # Get the integrated ZTD
        wet_total, hydro_total = np.zeros(wet.shape), np.zeros(hydro.shape)
        for level in range(wet.shape[2]):
            wet_total[..., level] = 1e-6 * np.trapz(
                wet[..., level:], x=self._zs[level:], axis=2
            )
            hydro_total[..., level] = 1e-6 * np.trapz(
                hydro[..., level:], x=self._zs[level:], axis=2
            )
        self._hydrostatic_ztd = hydro_total
        self._wet_ztd = wet_total

    @property
    def bbox(self) -> list:
        """Obtains the bounding box of the weather model in lat/lon CRS.

        Returns
        -------
        list
            xmin, ymin, xmax, ymax

        Raises
        ------
        ValueError
           When `self.files` is None.

        """
        if self._bbox is None:
            path_weather_model = self.out_file(self.get_wm_loc())
            if not Path(path_weather_model).exists():
                raise ValueError("Need to save cropped weather model as netcdf")

            with xarray.load_dataset(path_weather_model) as ds:
                if "x" in ds and "y" in ds:
                    xmin, xmax = ds.x.min(), ds.x.max()
                    ymin, ymax = ds.y.min(), ds.y.max()
                else:
                    xmin, xmax = ds.longitude.min(), ds.longitude.max()
                    ymin, ymax = ds.latitude.min(), ds.latitude.max()

            wm_proj = self._proj
            xs, ys = [xmin, xmin, xmax, xmax], [ymin, ymax, ymin, ymax]
            t = Transformer.from_crs(wm_proj.to_epsg(), 4326, always_xy=True)
            lons, lats = t.transform(xs, ys)

            ## projected weather models may not be aligned N/S
            ## should only matter for warning messages
            W, E = np.min(lons), np.max(lons)
            # S, N = np.sort([lats[np.argmin(lons)], lats[np.argmax(lons)]])
            S, N = np.min(lats), np.max(lats)
            self._bbox = W, S, E, N

        return self._bbox

    def _trim_extent(self, extent):
        """Get the bounding box around a set of lats/lons."""
        lat = self._lats.copy()
        lon = self._lons.copy()
        lat[np.isnan(lat)] = np.nanmean(lat)
        lon[np.isnan(lon)] = np.nanmean(lon)
        mask = (
            (lat >= extent[0])
            & (lat <= extent[1])
            & (lon >= extent[2])
            & (lon <= extent[3])
        )
        ma1 = np.sum(mask, axis=1).astype("bool")
        ma2 = np.sum(mask, axis=0).astype("bool")
        if np.sum(ma1) == 0 and np.sum(ma2) == 0:
            # Don't need to remove any points
            return

        # indices of the part of the grid to keep
        ny, nx, nz = self._p.shape
        index1 = max(np.arange(len(ma1))[ma1][0] - 2, 0)
        index2 = min(np.arange(len(ma1))[ma1][-1] + 2, ny)
        index3 = max(np.arange(len(ma2))[ma2][0] - 2, 0)
        index4 = min(np.arange(len(ma2))[ma2][-1] + 2, nx)

        # subset around points of interest
        self._lons = self._lons[index1:index2, index3:index4]
        self._lats = self._lats[index1:index2, index3:index4]
        self._xs = self._xs[index3:index4]
        self._ys = self._ys[index1:index2]
        self._p = self._p[index1:index2, index3:index4, ...]
        self._t = self._t[index1:index2, index3:index4, ...]
        self._e = self._e[index1:index2, index3:index4, ...]

        self._wet_refractivity = self._wet_refractivity[
            index1:index2, index3:index4, ...
        ]
        self._hydrostatic_refractivity = self._hydrostatic_refractivity[
            index1:index2, index3:index4, :
        ]

    def _calculategeoh(self, z, lnsp):
        """Calculate pressure, geopotential, and geopotential height.

        from the surface pressure and model levels provided by a weather model.
        The model levels are numbered from the highest eleveation to the lowest.
        Inputs:
            self - weather model object with parameters a, b defined
            z    - 3-D array of surface heights for the location(s) of interest
            lnsp - log of the surface pressure
        Outputs:
            geopotential - The geopotential in units of height times acceleration
            pressurelvs  - The pressure at each of the model levels for each of
                           the input points
            geoheight    - The geopotential heights.
        """
        return utils.calcgeoh(
            lnsp, self._t, self._q, z, self._a, self._b, self._R_d, self._levels
        )

    def _uniform_in_z(self, _zlevels=None):
        """Interpolate all variables to a regular grid in z."""
        nx, ny = self._p.shape[:2]

        # new regular z-spacing
        if _zlevels is None:
            try:
                _zlevels = self._zlevels
            except BaseException:
                _zlevels = np.nanmean(self._zs, axis=(0, 1))

        new_zs = np.tile(_zlevels, (nx, ny, 1))

        # re-assign values to the uniform z
        self._t = utils.interpolate_along_axis(
            self._zs, self._t, new_zs, axis=2
        ).astype(np.float32)
        self._p = utils.interpolate_along_axis(
            self._zs, self._p, new_zs, axis=2
        ).astype(np.float32)
        self._e = utils.interpolate_along_axis(
            self._zs, self._e, new_zs, axis=2
        ).astype(np.float32)

        self._zs = _zlevels
        self._xs = np.unique(self._xs)
        self._ys = np.unique(self._ys)

    def _check_for_nans(self):
        """Fill in NaN-values."""
        self._p = utils.fillna3d(self._p)
        self._t = utils.fillna3d(
            self._t, fill_value=1e16
        )  # to avoid division by zero later on
        self._e = utils.fillna3d(self._e)

    def out_file(self, outLoc):
        """Path to the output file."""
        s = np.floor(self._ll_bounds[0])
        S = f"{np.abs(s):.0f}S" if s < 0 else f"{s:.0f}N"

        n = np.ceil(self._ll_bounds[1])
        N = f"{np.abs(n):.0f}S" if n < 0 else f"{n:.0f}N"

        w = np.floor(self._ll_bounds[2])
        W = f"{np.abs(w):.0f}W" if w < 0 else f"{w:.0f}E"

        e = np.ceil(self._ll_bounds[3])
        E = f"{np.abs(e):.0f}W" if e < 0 else f"{e:.0f}E"
        out_name = f'{self._time.strftime("%Y_%m_%d_T%H_%M_%S")}_{S}_{N}_{W}_{E}.nc'
        f = f"{self._Name}_" + out_name
        return Path(outLoc) / f

    def filename(self, time=None, outLoc="weather_files"):
        """Create a filename to store the weather model."""
        Path(outLoc).mkdir(parents=True, exist_ok=True)

        if time is None:
            if self._time is None:
                raise ValueError(
                    "Time must be specified before the file can be written"
                )
            else:
                time = self._time

        f = Path(outLoc) / "{}_{}.{}".format(
            self._Name, datetime.datetime.strftime(time, "%Y_%m_%d_T%H_%M_%S"), "nc"
        )

        self.files = [f]
        return f

    def write(self):
        """By calling the abstract/modular netcdf writer.

        write the weather model data and refractivity to an
        NETCDF4 file that can be accessed by external programs.
        """
        # Generate the filename
        f = self._out_name

        attrs_dict = {
            "Conventions": "CF-1.6",
            "datetime": datetime.datetime.strftime(self._time, "%Y_%m_%dT%H_%M_%S"),
            "date_created": datetime.datetime.now().strftime("%Y_%m_%dT%H_%M_%S"),
            "title": "Weather model data and delay calculations",
            "model_name": self._Name,
        }

        dimension_dict = {
            "x": ("x", self._xs),
            "y": ("y", self._ys),
            "z": ("z", self._zs),
            "latitude": (("y", "x"), self._lats),
            "longitude": (("y", "x"), self._lons),
            "datetime_utc": self._time.replace(tzinfo=None),
        }

        dataset_dict = {
            "t": (("z", "y", "x"), self._t.swapaxes(0, 2).swapaxes(1, 2)),
            "p": (("z", "y", "x"), self._p.swapaxes(0, 2).swapaxes(1, 2)),
            "e": (("z", "y", "x"), self._e.swapaxes(0, 2).swapaxes(1, 2)),
            "wet": (
                ("z", "y", "x"),
                self._wet_refractivity.swapaxes(0, 2).swapaxes(1, 2),
            ),
            "hydro": (
                ("z", "y", "x"),
                self._hydrostatic_refractivity.swapaxes(0, 2).swapaxes(1, 2),
            ),
            "wet_total": (("z", "y", "x"), self._wet_ztd.swapaxes(0, 2).swapaxes(1, 2)),
            "hydro_total": (
                ("z", "y", "x"),
                self._hydrostatic_ztd.swapaxes(0, 2).swapaxes(1, 2),
            ),
        }

        ds = xarray.Dataset(
            data_vars=dataset_dict, coords=dimension_dict, attrs=attrs_dict
        )

        # Define units
        ds["t"].attrs["units"] = "K"
        ds["e"].attrs["units"] = "Pa"
        ds["p"].attrs["units"] = "Pa"
        ds["wet"].attrs["units"] = "dimentionless"
        ds["hydro"].attrs["units"] = "dimentionless"
        ds["wet_total"].attrs["units"] = "m"
        ds["hydro_total"].attrs["units"] = "m"

        # Define standard names
        ds["t"].attrs["standard_name"] = "temperature"
        ds["e"].attrs["standard_name"] = "humidity"
        ds["p"].attrs["standard_name"] = "pressure"
        ds["wet"].attrs["standard_name"] = "wet_refractivity"
        ds["hydro"].attrs["standard_name"] = "hydrostatic_refractivity"
        ds["wet_total"].attrs["standard_name"] = "total_wet_refractivity"
        ds["hydro_total"].attrs["standard_name"] = "total_hydrostatic_refractivity"

        # projection information
        ds["proj"] = int()
        for k, v in self._proj.to_cf().items():
            ds.proj.attrs[k] = v
        for var in ds.data_vars:
            ds[var].attrs["grid_mapping"] = "proj"

        # write to file and return the filename
        ds.to_netcdf(f)
        return f


class ECMWF(WeatherModel):
    """Implement ECMWF models."""

    def __init__(self, level_type="ml"):
        """Initiate ECMWF model."""
        # initialize a weather model
        WeatherModel.__init__(self)

        # model constants
        self._k1 = 0.776  # [K/Pa]
        self._k2 = 0.233  # [K/Pa]
        self._k3 = 3.75e3  # [K^2/Pa]

        self._time_res = TIME_RES["ECMWF"]

        self._lon_res = 0.25
        self._lat_res = 0.25
        self._proj = CRS.from_epsg(4326)

        # self._model_level_type = 'ml'  # Default
        self._Name = "ECMWF"
        self._dataset = "ecmwf"
        # self.__model_levels__()

        self._humidityType = "q"
        # Default, pressure levels are 'pl'

        self.set_level_type(level_type)

    def __pressure_levels__(self):
        self._zlevels = np.flipud(ml.LEVELS_25_HEIGHTS)
        self._levels = len(self._zlevels)

    def __model_levels__(self):
        self._levels = 137
        self._zlevels = np.flipud(ml.LEVELS_137_HEIGHTS)
        self._a = ml.A_137_HRES
        self._b = ml.B_137_HRES

    def load_weather(self, f=None):
        """Consistent class method to be implemented across all weather model types.

        As a result of calling this method, all of the variables (x, y, z, p, q,
        t, wet_refractivity, hydrostatic refractivity, e) should be fully
        populated.
        """
        f = self.files[0] if f is None else f
        self._load_model_level(f)

    def _load_model_level(self, fname):
        # read data from netcdf file

        lats, lons, xs, ys, t, q, lnsp, z = self._make_data_cubes(fname)

        # ECMWF appears to give me this backwards
        if lats[0] > lats[1]:
            z = z[::-1]
            lnsp = lnsp[::-1]
            t = t[:, ::-1]
            q = q[:, ::-1]
            lats = lats[::-1]
        # Lons is usually ok, but we'll throw in a check to be safe
        if lons[0] > lons[1]:
            z = z[..., ::-1]
            lnsp = lnsp[..., ::-1]
            t = t[..., ::-1]
            q = q[..., ::-1]
            lons = lons[::-1]
        # pyproj gets fussy if the latitude is wrong, plus our
        # interpolator isn't clever enough to pick up on the fact that
        # they are the same
        lons[lons > 180] -= 360

        self._t = t
        self._q = q
        geo_hgt, pres, hgt = self._calculategeoh(z, lnsp)

        self._lons, self._lats = np.meshgrid(lons, lats)

        # ys is latitude
        self._get_heights(self._lats, hgt.transpose(1, 2, 0))
        h = self._zs.copy()

        # We want to support both pressure levels and true pressure grids.
        # If the shape has one dimension, we'll scale it up to act as a
        # grid, otherwise we'll leave it alone.
        if len(pres.shape) == 1:
            self._p = np.broadcast_to(pres[:, np.newaxis, np.newaxis], self._zs.shape)
        else:
            self._p = pres

        # Re-structure everything from (heights, lats, lons) to (lons, lats, heights)
        self._p = self._p.transpose(1, 2, 0)
        self._t = self._t.transpose(1, 2, 0)
        self._q = self._q.transpose(1, 2, 0)

        # Flip all the axis so that zs are in order from bottom to top
        # lats / lons are simply replicated to all heights so they don't need flipped
        self._p = np.flip(self._p, axis=2)
        self._t = np.flip(self._t, axis=2)
        self._q = np.flip(self._q, axis=2)
        self._ys = self._lats.copy()
        self._xs = self._lons.copy()
        self._zs = np.flip(h, axis=2)

    def _load_pressure_level(self, filename, *args, **kwargs):
        with xarray.open_dataset(filename) as block:
            # Pull the data
            z = np.squeeze(block["z"].values)
            t = np.squeeze(block["t"].values)
            q = np.squeeze(block["q"].values)
            lats = np.squeeze(block.latitude.values)
            lons = np.squeeze(block.longitude.values)
            levels = np.squeeze(block.level.values) * 100

        z = np.flip(z, axis=1)

        # ECMWF appears to give me this backwards
        if lats[0] > lats[1]:
            z = z[::-1]
            t = t[:, ::-1]
            q = q[:, ::-1]
            lats = lats[::-1]
        # Lons is usually ok, but we'll throw in a check to be safe
        if lons[0] > lons[1]:
            z = z[..., ::-1]
            t = t[..., ::-1]
            q = q[..., ::-1]
            lons = lons[::-1]
        # pyproj gets fussy if the latitude is wrong, plus our
        # interpolator isn't clever enough to pick up on the fact that
        # they are the same
        lons[lons > 180] -= 360

        self._t = t
        self._q = q

        geo_hgt = (z / self._g0).transpose(1, 2, 0)

        # re-assign lons, lats to match heights
        self._lons, self._lats = np.meshgrid(lons, lats)

        # correct heights for latitude
        self._get_heights(self._lats, geo_hgt)

        self._p = np.broadcast_to(levels[np.newaxis, np.newaxis, :], self._zs.shape)

        # Re-structure from (heights, lats, lons) to (lons, lats, heights)
        self._t = self._t.transpose(1, 2, 0)
        self._q = self._q.transpose(1, 2, 0)
        self._ys = self._lats.copy()
        self._xs = self._lons.copy()

        # flip z to go from surface to toa
        self._p = np.flip(self._p, axis=2)
        self._t = np.flip(self._t, axis=2)
        self._q = np.flip(self._q, axis=2)

    def _make_data_cubes(self, fname):
        """Create a cube of data representing temperature and relative humidity.

        at specified pressure levels.
        """
        # get ll_bounds
        S, N, W, E = self._ll_bounds

        ds = xarray.open_dataset(fname)
        # ds = ds.assign_coords(longitude=(((ds.longitude + 180) % 360) - 180))
        ds = ds.assign_coords(longitude=(((ds.longitude) % 360) - 180))

        # mask based on query bounds
        m1 = (ds.latitude >= S) & (ds.latitude <= N)
        m2 = (ds.longitude >= W) & (ds.longitude <= E)
        block = ds.where(m1 & m2, drop=True)

        # Pull the data
        z = np.squeeze(block["z"].values)[0, ...]
        t = np.squeeze(block["t"].values)
        q = np.squeeze(block["q"].values)
        lnsp = np.squeeze(block["lnsp"].values)[0, ...]
        lats = np.squeeze(block.latitude.values)
        lons = np.squeeze(block.longitude.values)

        xs = lons.copy()
        ys = lats.copy()
        ds.close()

        if z.size == 0:
            raise RuntimeError(
                "There is no data in z, " "you may have a problem with your mask"
            )

        return lats, lons, xs, ys, t, q, lnsp, z


class HRES(ECMWF):
    """Implement HRES models."""

    def __init__(self, level_type="ml"):
        """Initiate HRES model."""
        # initialize a weather model
        WeatherModel.__init__(self)

        # model constants
        self._k1 = 0.776  # [K/Pa]
        self._k2 = 0.233  # [K/Pa]
        self._k3 = 3.75e3  # [K^2/Pa]

        # 9 km horizontal grid spacing. This is only used for extending
        # the download-buffer, i.e. not in subsequent processing.
        self._lon_res = 9.0 / 111  # 0.08108115
        self._lat_res = 9.0 / 111  # 0.08108115
        self._x_res = 9.0 / 111  # 0.08108115
        self._y_res = 9.0 / 111  # 0.08108115

        self._humidityType = "q"
        # Default, pressure levels are 'pl'
        self._expver = "1"
        self._classname = "od"
        self._dataset = "hres"
        self._Name = "HRES"
        self._proj = CRS.from_epsg(4326)

        self._time_res = TIME_RES[self._dataset.upper()]
        # Tuple of min/max years where data is available.
        self._valid_range = (
            datetime.datetime(1983, 4, 20).replace(
                tzinfo=datetime.timezone(offset=datetime.timedelta())
            ),
            datetime.datetime.now(datetime.timezone.utc),
        )
        # Availability lag time in days
        self._lag_time = datetime.timedelta(hours=6)

        self.set_level_type(level_type)

    def _update_a_b(self):
        # Before 2013-06-26, there were only 91 model levels.
        # The mapping coefficients below are extracted
        # based on https://www.ecmwf.int/en/forecasts/documentation-and-support/91-model-levels
        self._levels = 91
        self._zlevels = np.flipud(ml.LEVELS_91_HEIGHTS)
        self._a = ml.A_91_HRES
        self._b = ml.B_91_HRES

    def load_weather(self, f=None):
        """Consistent class method to be implemented across all weather model types.

        As a result of calling this method, all of the variables (x, y, z, p, q,
        t, wet_refractivity, hydrostatic refractivity, e) should be fully
        populated.
        """
        f = self.files[0] if f is None else f

        if self._model_level_type == "ml":
            if self._time < datetime.datetime(2013, 6, 26, 0, 0, 0):
                self._update_a_b()
            self._load_model_level(f)
        elif self._model_level_type == "pl":
            self._load_pressure_levels(f)



import numpy as np
from numpy.typing import ArrayLike


def goldstein(phase: ArrayLike, alpha: float, psize: int = 32) -> np.ndarray:
    """Apply the Goldstein adaptive filter to the given data.

    Parameters
    ----------
    phase : np.ndarray
        2D complex array containing the data to be filtered.
    alpha : float
        Filtering parameter for Goldstein algorithm
        Must be between 0 (no filtering) and 1 (maximum filtering)
    psize : int, optional
        edge length of square patch
        Default = 32

    Returns
    -------
        2D numpy array of filtered data.

    """

    def apply_pspec(data: ArrayLike):
        # NaN is allowed value
        if alpha < 0:
            raise ValueError(f"alpha must be >= 0, got {alpha = }")

        wgt = np.power(np.abs(data) ** 2, alpha / 2)
        data = wgt * data
        return data

    def make_wgt(nxp: int, nyp: int) -> np.ndarray:
        # Create arrays of horizontal and vertical weights
        wx = 1.0 - np.abs(np.arange(nxp // 2) - (nxp / 2.0 - 1.0)) / (nxp / 2.0 - 1.0)
        wy = 1.0 - np.abs(np.arange(nyp // 2) - (nyp / 2.0 - 1.0)) / (nyp / 2.0 - 1.0)
        # Compute the outer product of wx and wy to create
        # the top-left quadrant of the weight matrix
        quadrant = np.outer(wy, wx)
        # Create a full weight matrix by mirroring the quadrant along both axes
        wgt = np.block(
            [
                [quadrant, np.flip(quadrant, axis=1)],
                [np.flip(quadrant, axis=0), np.flip(np.flip(quadrant, axis=0), axis=1)],
            ]
        )
        return wgt

    def patch_goldstein_filter(
        data: ArrayLike, wgt: ArrayLike, psize: int
    ) -> np.ndarray:
        """Apply the filter to a single patch of data.

        Parameters
        ----------
        data : np.ndarray
            2D complex array containing the data to be filtered.
        wgt : np.ndarray
            weight matrix for summing neighboring data
        psize : int
            edge length of square FFT area

        Returns
        -------
            2D numpy array of filtered data.

        """
        # Calculate alpha
        data = np.fft.fft2(data, s=(psize, psize))
        data = apply_pspec(data)
        data = np.fft.ifft2(data, s=(psize, psize))
        return wgt * data

    def apply_goldstein_filter(data: ArrayLike) -> np.ndarray:
        # Create an empty array for the output
        out = np.zeros(data.shape, dtype=np.complex64)
        # ignore processing for empty chunks
        if np.all(np.isnan(data)):
            return data
        # Create the weight matrix
        wgt_matrix = make_wgt(psize, psize)
        # Iterate over windows of the data
        for i in range(0, data.shape[0] - psize, psize // 2):
            for j in range(0, data.shape[1] - psize, psize // 2):
                # Create proocessing windows
                data_window = data[i : i + psize, j : j + psize]
                wgt_window = wgt_matrix[: data_window.shape[0], : data_window.shape[1]]
                # Apply the filter to the window
                filtered_window = patch_goldstein_filter(data_window, wgt_window, psize)
                # Add the result to the output array
                slice_i = slice(i, min(i + psize, out.shape[0]))
                slice_j = slice(j, min(j + psize, out.shape[1]))
                out[slice_i, slice_j] += filtered_window[
                    : slice_i.stop - slice_i.start, : slice_j.stop - slice_j.start
                ]
        return out

    return apply_goldstein_filter(phase)



from __future__ import annotations

import datetime
import logging
from dataclasses import dataclass
from os import fspath
from pathlib import Path
from typing import Mapping, Sequence

import numpy as np
import opera_utils as oput
from opera_utils import get_dates, group_by_date
from osgeo import gdal
from rasterio.crs import CRS
from rasterio.warp import transform_bounds
from scipy.interpolate import RegularGridInterpolator

from dolphin import io
from dolphin._types import Bbox, Filename, TropoModel, TropoType
from dolphin.timeseries import ReferencePoint
from dolphin.utils import format_date_pair

from ._netcdf import delay_from_netcdf, group_netcdf_by_date

logger = logging.getLogger(__name__)

###########

__all__ = ["estimate_tropospheric_delay"]


@dataclass
class DelayParams:
    """Parameters for estimating tropospheric delay corrections."""

    x_coordinates: np.ndarray
    """Array of X coordinates."""

    y_coordinates: np.ndarray
    """Array of Y coordinates."""

    z_coordinates: np.ndarray
    """Array of Z coordinates."""

    SNWE: tuple[float, float, float, float]
    """ Bounding box of the data in SNWE format of RAiDER/PYAPS."""

    shape: tuple[int, int]
    """Shape of the data, specified as a tuple of two integers."""

    epsg: int
    """EPSG code for the coordinate reference system."""

    geotransform: list[float]
    """Sequence of geotransformation parameters."""

    tropo_model: TropoModel
    """Model used for tropospheric correction."""

    delay_type: str
    """Type of tropospheric delay."""

    reference_file: Filename
    """Filename used as reference files."""

    secondary_file: Filename
    """Filename used as secondary files."""

    interferogram: str
    """Identifier for the interferogram."""

    reference_time: datetime.datetime
    """The reference image time."""

    secondary_time: datetime.datetime
    """The secondary image time."""


def estimate_tropospheric_delay(
    ifg_file_list: Sequence[Path],
    slc_files: Mapping[tuple[datetime.datetime], Sequence[Filename]],
    troposphere_files: Sequence[Filename],
    geom_files: dict[str, Path],
    reference_point: ReferencePoint | None,
    output_dir: Path,
    file_date_fmt: str,
    tropo_model: TropoModel,
    tropo_delay_type: TropoType,
    epsg: int,
    bounds: Bbox,
) -> list[Path]:
    """Estimate the tropospheric delay corrections (in meters) for each interferogram.

    Parameters
    ----------
    ifg_file_list : Sequence[Path]
        List of interferogram files.
    slc_files : Mapping[tuple[datetime.datetime], Sequence[Filename]]
        Dictionary of SLC files indexed by date.
    troposphere_files : Sequence[Filename]
        list of troposphere files.
    geom_files : dict[str, Path]
        Dictionary of geometry files with height and incidence angle, or
        los_east and los_north.
    reference_point : tuple[int, int], optional
        Reference point (row, col) used during time series inversion.
        If not provided, no spatial referencing is performed.
    output_dir : Path
        Output directory.
    file_date_fmt : str
        Format of the date in the filename.
        Default is [dolphin.DEFAULT_DATETIME_FORMAT][]
    tropo_model : TropoModel
        Tropospheric model (ERA5, HRES, ...).
    tropo_delay_type : TropoType
        Tropospheric delay type ('wet', 'hydrostatic', 'comb').
    epsg : int
        the EPSG code of the input data
    bounds : Bbox
        Output bounds.

    Returns
    -------
    list[Path]
        List of newly created tropospheric phase delay geotiffs.
        Units are in meters.

    """
    # Read geogrid data
    xsize, ysize = io.get_raster_xysize(ifg_file_list[0])
    gt = io.get_raster_gt(ifg_file_list[0])
    ycoord, xcoord = oput.create_yx_arrays(gt, (ysize, xsize))  # 500 m spacing

    if epsg != 4326:
        left, bottom, right, top = transform_bounds(
            CRS.from_epsg(epsg), CRS.from_epsg(4326), *bounds
        )
    else:
        left, bottom, right, top = bounds

    tropo_height_levels = np.concatenate(([-100], np.arange(0, 9000, 500)))

    delay_type = tropo_delay_type.value

    if str(troposphere_files[0]).endswith(".nc"):
        tropo_files = group_netcdf_by_date(troposphere_files)
        tropo_run = compute_tropo_delay_from_netcdf
        if tropo_delay_type.value == "hydrostatic":
            delay_type = "hydro"
    else:
        tropo_files = group_by_date(troposphere_files, file_date_fmt=file_date_fmt)
        tropo_run = compute_pyaps
        if tropo_delay_type.value == "hydrostatic":
            delay_type = "dry"

    tropo_date_list = [x.date() for x in tropo_files]

    output_tropo_dir = output_dir / "troposphere"
    output_tropo_dir.mkdir(exist_ok=True)
    output_paths: list[Path] = []
    for ifg in ifg_file_list:
        ref_date, sec_date = get_dates(ifg)[0:2]

        date_str = format_date_pair(ref_date, sec_date)
        name = f"{date_str}_tropoDelay_{tropo_model.value}_LOS_{delay_type}.tif"
        tropo_delay_product_path = output_tropo_dir / name
        output_paths.append(tropo_delay_product_path)

        if tropo_delay_product_path.exists():
            logger.info(f"{tropo_delay_product_path} exists, skipping")
            continue

        reference_date = next(key for key in slc_files if ref_date in key)
        # temporary fix for compressed SLCs while the required metadata i not included
        # in them. there will be modification in a future PR to add metadata to
        # compressed SLCs
        secondary_date = next(
            key
            for key in slc_files
            if sec_date in key and "compressed" not in str(slc_files[key][0]).lower()
        )

        if (
            ref_date.date() not in tropo_date_list
            or sec_date.date() not in tropo_date_list
        ):
            logger.warning(f"Weather-model files do not exist for {date_str}, skipping")
            continue

        secondary_time = oput.get_zero_doppler_time(slc_files[secondary_date][0])
        if "compressed" in str(slc_files[reference_date][0]).lower():
            # this is for when we have compressed slcs but the actual
            # reference date does not exist in the input data
            reference_time = datetime.datetime.combine(
                reference_date[0].date(), secondary_time.time()
            )
        else:
            reference_time = oput.get_zero_doppler_time(slc_files[reference_date][0])

        if str(troposphere_files[0]).endswith(".nc"):
            ref_date = ref_date + datetime.timedelta(hours=reference_time.hour)
            sec_date = sec_date + datetime.timedelta(hours=secondary_time.hour)

        closest_date_ref = min(tropo_files.keys(), key=lambda x: abs(x - ref_date))
        closest_date_sec = min(tropo_files.keys(), key=lambda x: abs(x - sec_date))

        delay_parameters = DelayParams(
            x_coordinates=xcoord,
            y_coordinates=ycoord,
            z_coordinates=tropo_height_levels,
            SNWE=(bottom, top, left, right),
            epsg=epsg,
            tropo_model=tropo_model,
            delay_type=delay_type,
            shape=(ysize, xsize),
            geotransform=gt,
            reference_file=tropo_files[closest_date_ref][0],
            secondary_file=tropo_files[closest_date_sec][0],
            reference_time=reference_time,
            secondary_time=secondary_time,
            interferogram=format_date_pair(ref_date, sec_date),
        )

        logger.info(f"Running troposphere computation for {ref_date}, {sec_date}")
        delay_datacube = tropo_run(delay_parameters)

        tropo_delay_2d = compute_2d_delay(delay_parameters, delay_datacube, geom_files)

        if reference_point is not None:
            ref_row, ref_col = reference_point
            tropo_delay_2d -= tropo_delay_2d[ref_row, ref_col]

        # Write 2D tropospheric correction layer to disc
        io.write_arr(
            arr=tropo_delay_2d,
            output_name=tropo_delay_product_path,
            like_filename=ifg,
            units="meters",
        )

    return output_paths


def compute_pyaps(delay_parameters: DelayParams) -> np.ndarray:
    """Compute tropospheric delay datacube using PyAPS.

    Parameters
    ----------
    delay_parameters : DelayParams
        delay parameters and grid information.

    Returns
    -------
    np.ndarray
        tropospheric delay datacube.

    """
    import pyaps3 as pa

    # X and y for the entire datacube
    y_2d, x_2d = np.meshgrid(
        delay_parameters.y_coordinates, delay_parameters.x_coordinates, indexing="ij"
    )

    # Lat/lon coordinates
    lat_datacube, lon_datacube = oput.transform_xy_to_latlon(
        delay_parameters.epsg, x_2d, y_2d
    )

    tropo_delay_datacube_list = []

    for hgt in delay_parameters.z_coordinates:
        dem_datacube = np.full(lat_datacube.shape, hgt)

        # Delay for the reference image
        ref_aps_estimator = pa.PyAPS(
            fspath(delay_parameters.reference_file),
            dem=dem_datacube,
            inc=0.0,
            lat=lat_datacube,
            lon=lon_datacube,
            grib=delay_parameters.tropo_model,
            humidity="Q",
            model=delay_parameters.tropo_model,
            verb=False,
            Del=delay_parameters.delay_type,
        )

        phs_ref = ref_aps_estimator.getdelay()

        # Delay for the secondary image
        second_aps_estimator = pa.PyAPS(
            fspath(delay_parameters.secondary_file),
            dem=dem_datacube,
            inc=0.0,
            lat=lat_datacube,
            lon=lon_datacube,
            grib=delay_parameters.tropo_model,
            humidity="Q",
            model=delay_parameters.tropo_model,
            verb=False,
            Del=delay_parameters.delay_type,
        )

        phs_second = second_aps_estimator.getdelay()

        # Convert the delay in meters to radians
        relative_delay = phs_second - phs_ref
        tropo_delay_datacube_list.append(relative_delay)

    # Tropo delay datacube
    tropo_delay_datacube = np.stack(tropo_delay_datacube_list)
    # Create a maksed datacube that excludes the NaN values
    return np.ma.masked_invalid(tropo_delay_datacube)


def compute_tropo_delay_from_netcdf(delay_parameters: DelayParams) -> np.ndarray:
    """Compute tropospheric delay (in meters) datacube from netcdf tropo file.

    Parameters
    ----------
    delay_parameters : DelayParams
        delay parameters and grid information.

    Returns
    -------
    np.ndarray
        tropospheric delay datacube.

    """
    reference_weather_model_file = delay_parameters.reference_file
    secondary_weather_model_file = delay_parameters.secondary_file

    # Troposphere delay datacube computation
    tropo_delay_reference = delay_from_netcdf(
        dt=delay_parameters.reference_time,
        weather_model_file=reference_weather_model_file,
        SNWE=delay_parameters.SNWE,
        height_levels=delay_parameters.z_coordinates,
        out_proj=delay_parameters.epsg,
        weather_model=delay_parameters.tropo_model,
    )

    tropo_delay_secondary = delay_from_netcdf(
        dt=delay_parameters.secondary_time,
        weather_model_file=secondary_weather_model_file,
        SNWE=delay_parameters.SNWE,
        height_levels=delay_parameters.z_coordinates,
        out_proj=delay_parameters.epsg,
        weather_model=delay_parameters.tropo_model,
    )

    # comb is the summation of wet and hydro components
    if delay_parameters.delay_type == "comb":
        tropo_delay = (
            tropo_delay_reference["wet"]
            + tropo_delay_reference["hydro"]
            - tropo_delay_secondary["wet"]
            - tropo_delay_secondary["hydro"]
        )
    else:
        tropo_delay = (
            tropo_delay_reference[delay_parameters.delay_type]
            - tropo_delay_secondary[delay_parameters.delay_type]
        )

    # Convert it to convention where positive means toward the satellite
    tropo_delay_datacube = -1 * tropo_delay

    # Create a masked datacube that excludes the NaN values
    tropo_delay_datacube_masked = np.ma.masked_invalid(tropo_delay_datacube)

    # Interpolate to radar grid to keep its dimension consistent with other datacubes
    tropo_delay_interpolator = RegularGridInterpolator(
        (tropo_delay_reference.z, tropo_delay_reference.y, tropo_delay_reference.x),
        tropo_delay_datacube_masked,
        method="linear",
        bounds_error=False,
    )

    # Interpolate the troposphere delay
    hv, yv, xv = np.meshgrid(
        delay_parameters.z_coordinates,
        delay_parameters.y_coordinates,
        delay_parameters.x_coordinates,
        indexing="ij",
    )

    pnts = np.stack((hv.flatten(), yv.flatten(), xv.flatten()), axis=-1)

    # Interpolate
    return tropo_delay_interpolator(pnts).reshape(hv.shape)


def compute_2d_delay(
    delay_parameters: DelayParams,
    delay_datacube: np.ndarray,
    geo_files: dict[str, Path],
) -> np.ndarray:
    """Compute 2D delay.

    Parameters
    ----------
    delay_parameters : DelayParams
        dataclass containing tropospheric delay data.

    delay_datacube : np.ndarray
        delay datacube for the x,y,z coordinates in delay_parameters

    geo_files : dict[str, Path]
        Dictionary containing paths to geospatial files.

    Returns
    -------
    np.ndarray
        Computed 2D delay in meters.

    """
    dem_file = Path(geo_files["height"])

    ysize, xsize = delay_parameters.shape
    x_origin, x_res, x_, y_origin, y_, y_res = delay_parameters.geotransform

    x = 0
    y = 0
    left = x_origin + x * x_res + y * x_
    top = y_origin + x * y_ + y * y_res

    x = xsize
    y = ysize

    right = x_origin + x * x_res + y * x_
    bottom = y_origin + x * y_ + y * y_res

    bounds = (left, bottom, right, top)

    crs = CRS.from_epsg(delay_parameters.epsg)

    options = gdal.WarpOptions(
        dstSRS=crs,
        format="MEM",
        xRes=x_res,
        yRes=y_res,
        outputBounds=bounds,
        outputBoundsSRS=crs,
        resampleAlg="near",
    )
    # Use the same suffix for warping
    target_ds = gdal.Warp(
        "",  # Output to memory to read
        fspath(dem_file.resolve()),
        options=options,
    )

    dem = target_ds.ReadAsArray()

    if "los_east" in geo_files:
        # ISCE3 geocoded products
        los_east = io.load_gdal(geo_files["los_east"])
        los_north = io.load_gdal(geo_files["los_north"])
        los_up = np.sqrt(1 - los_east**2 - los_north**2)
    else:
        # ISCE2 radar coordinate
        los_up = np.cos(np.deg2rad(io.load_gdal(geo_files["incidence_angle"])))

    mask = los_east > 0

    # Make the x/y arrays
    # Note that these are the center of the pixels, whereas the GeoTransform
    # is the upper left corner of the top left pixel.
    y = np.arange(y_origin, y_origin + y_res * ysize, y_res)
    x = np.arange(x_origin, x_origin + x_res * xsize, x_res)

    yv, xv = np.meshgrid(y, x, indexing="ij")

    tropo_delay_interpolator = RegularGridInterpolator(
        (
            delay_parameters.z_coordinates,
            delay_parameters.y_coordinates,
            delay_parameters.x_coordinates,
        ),
        delay_datacube,
        method="linear",
        bounds_error=False,
    )

    tropo_delay_2d = np.zeros(dem.shape, dtype=np.float32)

    nline = 100
    for i in range(0, dem.shape[1], 100):
        if i + 100 > dem.shape[0]:
            nline = dem.shape[0] - i
        pnts = np.stack(
            (
                dem[i : i + 100, :].flatten(),
                yv[i : i + 100, :].flatten(),
                xv[i : i + 100, :].flatten(),
            ),
            axis=-1,
        )
        tropo_delay_2d[i : i + 100, :] = tropo_delay_interpolator(pnts).reshape(
            nline, dem.shape[1]
        )

    return (tropo_delay_2d / los_up) * mask



import numpy as np
import pytest
from numpy.testing import assert_allclose
from osgeo import gdal

import dolphin.ps
from dolphin import io


def test_ps_block(slc_stack):
    # Run the PS selector on entire stack
    amp_mean, amp_disp, ps_pixels = dolphin.ps.calc_ps_block(
        np.abs(slc_stack),
        amp_dispersion_threshold=0.25,  # should be too low for random data
    )
    assert amp_mean.shape == amp_disp.shape == ps_pixels.shape
    assert amp_mean.dtype == amp_disp.dtype == np.float32
    assert ps_pixels.dtype == bool

    assert ps_pixels.sum() == 0

    assert amp_mean.min() > 0
    assert amp_disp.min() >= 0


def test_ps_nodata(slc_stack):
    s_nan = slc_stack.copy()
    s_nan[:, 0, 0] = np.nan
    # Run the PS selector on entire stack
    amp_mean, amp_disp, ps_pixels = dolphin.ps.calc_ps_block(
        np.abs(s_nan),
        amp_dispersion_threshold=0.95,  # high thresh shouldn't matter for nodata
    )
    assert amp_mean[0, 0] == 0
    assert amp_disp[0, 0] == 0
    assert not ps_pixels[0, 0]


def test_ps_threshold(slc_stack):
    _, _, ps_pixels = dolphin.ps.calc_ps_block(
        np.abs(slc_stack),
        amp_dispersion_threshold=100000,
    )
    assert ps_pixels.sum() == ps_pixels.size
    _, _, ps_pixels = dolphin.ps.calc_ps_block(
        np.abs(slc_stack),
        amp_dispersion_threshold=0,
    )
    assert ps_pixels.sum() == 0


@pytest.fixture()
def vrt_stack(tmp_path, slc_file_list):
    vrt_file = tmp_path / "test.vrt"
    return io.VRTStack(slc_file_list, outfile=vrt_file)


def test_create_ps(tmp_path, vrt_stack):
    ps_mask_file = tmp_path / "ps_pixels.tif"

    amp_dispersion_file = tmp_path / "amp_disp.tif"
    amp_mean_file = tmp_path / "amp_mean.tif"
    dolphin.ps.create_ps(
        reader=vrt_stack,
        like_filename=vrt_stack.outfile,
        output_amp_dispersion_file=amp_dispersion_file,
        output_amp_mean_file=amp_mean_file,
        output_file=ps_mask_file,
    )
    assert io.get_raster_dtype(ps_mask_file) == np.uint8
    assert io.get_raster_dtype(amp_mean_file) == np.float32
    assert io.get_raster_dtype(amp_dispersion_file) == np.float32


@pytest.fixture()
def vrt_stack_with_nans(tmp_path, raster_with_nan_block):
    vrt_file = tmp_path / "test_with_nans.vrt"
    return io.VRTStack([raster_with_nan_block, raster_with_nan_block], outfile=vrt_file)


def _write_zeros(file, shape):
    driver = gdal.GetDriverByName("ENVI")
    ds = driver.Create(str(file), shape[1], shape[0], 1, gdal.GDT_Float32)
    bnd = ds.GetRasterBand(1)
    bnd.WriteArray(np.zeros(shape))
    ds.SetMetadataItem("N", "0", "ENVI")
    ds = bnd = None


def test_multilook_ps_file(tmp_path, vrt_stack):
    ps_mask_file = tmp_path / "ps_pixels.tif"

    amp_dispersion_file = tmp_path / "amp_disp.tif"
    amp_mean_file = tmp_path / "amp_mean.tif"
    dolphin.ps.create_ps(
        reader=vrt_stack,
        like_filename=vrt_stack.outfile,
        output_amp_dispersion_file=amp_dispersion_file,
        output_amp_mean_file=amp_mean_file,
        output_file=ps_mask_file,
    )
    output_ps_file, output_amp_disp_file = dolphin.ps.multilook_ps_files(
        strides={"x": 5, "y": 3},
        ps_mask_file=ps_mask_file,
        amp_dispersion_file=amp_dispersion_file,
    )
    assert io.get_raster_dtype(output_ps_file) == np.uint8
    assert io.get_raster_dtype(output_amp_disp_file) == np.float32


def test_compute_combined_amplitude_means():
    # Test basic functionality
    amplitudes = np.array([[[1.0, 1.0], [1.0, 1.0]], [[6.0, 6.0], [11.0, 21.0]]])
    N = np.array([9, 1])
    expected = np.array([[1.5, 1.5], [2.0, 3.0]])
    result = dolphin.ps.combine_means(amplitudes, N)
    assert_allclose(result, expected, rtol=1e-5)

    #  Test with multiple groups
    amplitudes = np.random.randn(10, 2, 2) ** 2
    amp_mean_1 = np.mean(amplitudes[:5], axis=0)
    amp_mean_2 = np.mean(amplitudes[5:9], axis=0)
    amp_3 = amplitudes[9]
    result = dolphin.ps.combine_means(
        np.stack([amp_mean_1, amp_mean_2, amp_3]), [5, 4, 1]
    )
    assert_allclose(result, np.mean(amplitudes, axis=0), rtol=1e-5)

    # Test with all equal weights
    expected_equal = np.mean(amplitudes, axis=0)
    result_equal = dolphin.ps.combine_means(amplitudes, np.ones(len(amplitudes)))
    assert_allclose(result_equal, expected_equal, rtol=1e-5)


def test_compute_combined_amplitude_dispersions():
    # Test basic functionality

    amplitudes = np.random.randn(10, 2, 2) ** 2

    amp_mean, amp_disp, _ = dolphin.ps.calc_ps_block(amplitudes)

    N = [5, 4, 1]

    amp_mean_1, amp_disp_1, _ = dolphin.ps.calc_ps_block(amplitudes[:5])
    amp_mean_2, amp_disp_2, _ = dolphin.ps.calc_ps_block(amplitudes[5:9])

    mean_inputs = np.stack([amp_mean_1, amp_mean_2, amplitudes[9]])
    # Note: a dispersion of N=1 isn't really defined. we dont use that
    disp_inputs = np.stack([amp_disp_1, amp_disp_2, np.zeros_like(amplitudes[9])])

    combined_disp, combined_mean = dolphin.ps.combine_amplitude_dispersions(
        dispersions=disp_inputs, means=mean_inputs, N=N
    )
    assert_allclose(combined_disp, amp_disp, rtol=1e-5)


def test_single_group():
    """Test with a group where all N=1 (meaning we passed in just the amplitudes)."""
    amplitudes = np.random.randn(10, 2, 2) ** 2
    amp_mean, amp_disp, _ = dolphin.ps.calc_ps_block(amplitudes)
    N = [1] * len(amplitudes)
    result = dolphin.ps.combine_means(amplitudes, N)
    assert_allclose(result, amp_mean, rtol=1e-5)

    result_disp, result_mean = dolphin.ps.combine_amplitude_dispersions(
        np.zeros_like(amplitudes), amplitudes, N
    )
    assert_allclose(result_disp, amp_disp, rtol=1e-5)
    assert_allclose(result_mean, amp_mean, rtol=1e-5)



from __future__ import annotations

import datetime
import sys
from enum import Enum
from os import PathLike
from typing import (
    TYPE_CHECKING,
    NamedTuple,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

# Some classes are declared as generic in stubs, but not at runtime.
# In Python 3.9 and earlier, os.PathLike is not subscriptable, results in runtime error
if TYPE_CHECKING:
    from builtins import ellipsis

    Index = ellipsis | slice | int
    PathLikeStr = PathLike[str]
else:
    PathLikeStr = PathLike


class Bbox(NamedTuple):
    """Bounding box named tuple, defining extent in cartesian coordinates.

    Usage:

        Bbox(left, bottom, right, top)

    Attributes
    ----------
    left : float
        Left coordinate (xmin)
    bottom : float
        Bottom coordinate (ymin)
    right : float
        Right coordinate (xmax)
    top : float
        Top coordinate (ymax)

    """

    left: float
    bottom: float
    right: float
    top: float


class Strides(NamedTuple):
    """Decimation/striding factor in the y (row) and x (column) directions."""

    y: int
    x: int


class HalfWindow(NamedTuple):
    """Half-window size in the y (row) and x (column) directions."""

    y: int
    x: int


# Used for callable types
T = TypeVar("T")
P = ParamSpec("P")

DateOrDatetime = Union[datetime.datetime, datetime.date]


class ReferencePoint(NamedTuple):
    row: int
    col: int


class TropoModel(str, Enum):
    """Enumeration representing different tropospheric models."""

    ECMWF = "ECMWF"
    ERA5 = "ERA5"
    HRES = "HRES"
    ERAINT = "ERAINT"
    ERAI = "ERAI"
    MERRA = "MERRA"
    NARR = "NARR"
    HRRR = "HRRR"
    GMAO = "GMAO"


class TropoType(str, Enum):
    """Type of tropospheric delay."""

    WET = "wet"
    """Wet tropospheric delay."""
    DRY = "dry"
    """Dry delay (same as hydrostatic, named "dry" in PyAPS)"""
    HYDROSTATIC = "hydrostatic"
    """Hydrostatic (same as dry, named differently in raider)"""
    COMB = "comb"
    """Combined wet + dry delay."""


@runtime_checkable
class GeneralPath(Protocol):
    """A protocol to handle paths that can be either local or S3 paths."""

    def parent(self): ...

    def suffix(self): ...

    def resolve(self): ...

    def exists(self): ...

    def read_text(self): ...

    def __truediv__(self, other): ...

    def __str__(self) -> str: ...

    def __fspath__(self) -> str:
        return str(self)


PathOrStr = Union[str, PathLikeStr, GeneralPath]
Filename = PathOrStr  # May add a deprecation notice for `Filename`
# TypeVar added for generic functions which should return the same type as the input
PathLikeT = TypeVar("PathLikeT", str, PathLikeStr, GeneralPath)



"""Configuration and validation models for (sub)workflows."""

from ._common import *
from ._displacement import *
from ._enums import *
from ._ps import *
from ._unwrap_options import *
from ._unwrapping import *
from ._yaml_model import *



import warnings

import numpy as np
from numpy.typing import ArrayLike

from dolphin.utils import upsample_nearest


def compress(
    slc_stack: ArrayLike, pl_cpx_phase: ArrayLike, slc_mean: ArrayLike | None = None
):
    """Compress the stack of SLC data using the estimated phase.

    Parameters
    ----------
    slc_stack : ArrayLike
        The stack of complex SLC data, shape (nslc, rows, cols)
    pl_cpx_phase : ArrayLike
        The estimated complex phase from phase linking.
        shape  = (nslc, rows // strides.y, cols // strides.x)
    slc_mean : ArrayLike, optional
        The mean SLC magnitude, shape (rows, cols), to use as output pixel magnitudes.
        If None, the mean is computed from the input SLC stack.

    Returns
    -------
    np.array
        The compressed SLC data, shape (rows, cols)

    """
    # If the output is downsampled, we need to make `pl_cpx_phase` the same shape
    # as the output
    pl_estimate_upsampled = upsample_nearest(pl_cpx_phase, slc_stack.shape[1:])
    # For each pixel, project the SLCs onto the (normalized) estimated phase
    # by performing a pixel-wise complex dot product
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="invalid value encountered")
        phase = np.angle(
            np.nansum(slc_stack * np.conjugate(pl_estimate_upsampled), axis=0)
        )
    if slc_mean is None:
        slc_mean = np.mean(np.abs(slc_stack), axis=0)
    # If the phase is invalid, set the mean to NaN
    slc_mean[phase == 0] = np.nan
    return slc_mean * np.exp(1j * phase)



from __future__ import annotations

import itertools
import logging
import shutil
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
from tqdm.auto import tqdm

from dolphin import goldstein, interpolate, io
from dolphin._types import Filename
from dolphin.utils import DummyProcessPoolExecutor, full_suffix
from dolphin.workflows import UnwrapMethod, UnwrapOptions

from ._constants import (
    CONNCOMP_SUFFIX,
    CONNCOMP_SUFFIX_ZEROED,
    DEFAULT_CCL_NODATA,
    DEFAULT_UNW_NODATA,
    UNW_SUFFIX,
    UNW_SUFFIX_ZEROED,
)
from ._snaphu_py import grow_conncomp_snaphu, unwrap_snaphu_py
from ._tophu import multiscale_unwrap
from ._unwrap_3d import unwrap_spurt
from ._utils import create_combined_mask, set_nodata_values
from ._whirlwind import unwrap_whirlwind

logger = logging.getLogger(__name__)

__all__ = ["run", "unwrap"]

DEFAULT_OPTIONS = UnwrapOptions()


def run(
    ifg_filenames: Sequence[Filename],
    cor_filenames: Sequence[Filename],
    output_path: Filename,
    *,
    unwrap_options: UnwrapOptions = DEFAULT_OPTIONS,
    nlooks: float = 5,
    temporal_coherence_file: Filename | None = None,
    mask_filename: Filename | None = None,
    unw_nodata: float | None = DEFAULT_UNW_NODATA,
    ccl_nodata: int | None = DEFAULT_CCL_NODATA,
    scratchdir: Filename | None = None,
    delete_intermediate: bool = True,
    overwrite: bool = False,
) -> tuple[list[Path], list[Path]]:
    """Run snaphu on all interferograms in a directory.

    Parameters
    ----------
    ifg_filenames : Sequence[Filename]
        Paths to input interferograms.
    cor_filenames : Sequence[Filename]
        Paths to input correlation files. Order must match `ifg_filenames`.
    output_path : Filename
        Path to output directory.
    unwrap_options : UnwrapOptions, optional
        [`UnwrapOptions`][dolphin.workflows.config.UnwrapOptions] config object
        with parameters and settings for unwrapping.
    nlooks : int, optional
        Effective number of looks used to form the input correlation data.
    temporal_coherence_file : Filename, optional
        Path to temporal coherence file from phase linking.
    mask_filename : Filename, optional
        Path to binary byte mask file, by default None.
        Assumes that 1s are valid pixels and 0s are invalid.
    unw_nodata : float , optional.
        Requested nodata value for the unwrapped phase.
        Default = 0
    ccl_nodata : float, optional
        Requested nodata value for connected component labels.
        Default = max value of UInt16 (65535)
    scratchdir : Filename, optional
        Path to scratch directory to hold intermediate files.
        If None, uses the unwrapper's default.
    delete_intermediate : bool, default = True
        Delete the temporary files made in the scratchdir after completion.
        If True, will make separate folders inside `scratchdir` for cleaner
        removals (in case `scratchdir`) has other contents.
        Must specify `scratchdir` for this option to be used.
    overwrite : bool, optional, default = False
        Overwrite existing unwrapped files.

    Returns
    -------
    unw_paths : list[Path]
        list of unwrapped files names
    conncomp_paths : list[Path]
        list of connected-component-label files names

    """
    if scratchdir is None:
        delete_intermediate = False

    if len(cor_filenames) != len(ifg_filenames):
        msg = (
            "Number of correlation files does not match number of interferograms."
            f" Found {len(cor_filenames)} correlation files and"
            f" {len(ifg_filenames)} interferograms."
        )
        raise ValueError(msg)

    output_path = Path(output_path)
    if unwrap_options.unwrap_method == UnwrapMethod.SPURT:
        if temporal_coherence_file is None:
            # TODO: we should make this a mask, instead of requiring this.
            # we'll need to change spurt
            raise ValueError("temporal coherence required for spurt unwrapping")
        unw_paths, conncomp_paths = unwrap_spurt(
            ifg_filenames=ifg_filenames,
            output_path=output_path,
            temporal_coherence_file=temporal_coherence_file,
            cor_filenames=cor_filenames,
            mask_filename=mask_filename,
            options=unwrap_options.spurt_options,
            scratchdir=scratchdir,
        )
        return unw_paths, conncomp_paths

    ifg_suffixes = [full_suffix(f) for f in ifg_filenames]
    all_out_files = [
        (output_path / Path(f).name.replace(suf, UNW_SUFFIX))
        for f, suf in zip(ifg_filenames, ifg_suffixes)
    ]
    in_files, out_files = [], []
    for inf, outf in zip(ifg_filenames, all_out_files):
        if Path(outf).exists() and not overwrite:
            logger.info(f"{outf} exists. Skipping.")
            continue

        in_files.append(inf)
        out_files.append(outf)
    logger.info(f"{len(out_files)} left to unwrap")

    if mask_filename:
        mask_filename = Path(mask_filename).resolve()

    if delete_intermediate:
        assert scratchdir is not None  # can't be none from the previous check
        scratch_dirs: list[Path | None] = [
            Path(scratchdir) / f"scratch-{Path(ifg_file).stem}" for ifg_file in in_files
        ]
    else:
        scratch_dirs = itertools.repeat(scratchdir)  # type: ignore[assignment]
    # This keeps it from spawning a new process for a single job.
    max_jobs = unwrap_options.n_parallel_jobs

    Executor = ThreadPoolExecutor if max_jobs > 1 else DummyProcessPoolExecutor
    with Executor(max_workers=max_jobs) as exc:
        futures = [
            exc.submit(
                unwrap,
                ifg_filename=ifg_file,
                corr_filename=cor_file,
                unw_filename=out_file,
                nlooks=nlooks,
                mask_filename=mask_filename,
                unwrap_options=unwrap_options,
                unw_nodata=unw_nodata,
                ccl_nodata=ccl_nodata,
                scratchdir=cur_scratch,
                delete_scratch=delete_intermediate,
            )
            for ifg_file, out_file, cor_file, cur_scratch in zip(
                in_files, out_files, cor_filenames, scratch_dirs
            )
        ]
        for fut in tqdm(as_completed(futures)):
            # We're not passing all the unw files in, so we need to tally up below
            _unw_path, _cc_path = fut.result()

    if unwrap_options.zero_where_masked and mask_filename is not None:
        all_out_files = [
            Path(str(outf).replace(UNW_SUFFIX, UNW_SUFFIX_ZEROED))
            for outf in all_out_files
        ]
        conncomp_files = [
            Path(str(outf).replace(UNW_SUFFIX_ZEROED, CONNCOMP_SUFFIX_ZEROED))
            for outf in all_out_files
        ]
    else:
        conncomp_files = [
            Path(str(outf).replace(UNW_SUFFIX, CONNCOMP_SUFFIX))
            for outf in all_out_files
        ]
    return all_out_files, conncomp_files


def transfer_ambiguities(wrapped: np.ndarray, unw_est: np.ndarray) -> np.ndarray:
    """Compute unwrapped phase by transferring ambiguities from an unwrapped estimate.

    Transfer the ambiguities from an unwrapped phase estimate to the original wrapped
    phase in order to form a new unwrapped phase array that is congruent (i.e. differs
    from the wrapped phase only by multiples of 2pi).

    Parameters
    ----------
    wrapped : numpy.ndarray
        The initial wrapped phase array, in radians. A 2-D, real-valued array.
    unw_est : numpy.ndarray
        An estimate of the unwrapped phase, in radians. A 2-D, real-valued array with
        the same shape as `wrapped`. May differ from the wrapped phase by non-integer
        cycles.

    Returns
    -------
    numpy.ndarray
        The unwrapped phase data, rounded pixel-wise to the nearest value that differs
        from the wrapped phase by a multiple of 2pi.

    """
    # Measure the difference between the unwrapped & wrapped phase, rounded to the
    # nearest phase cycle.
    ambiguity = np.round((unw_est - wrapped) / (2 * np.pi))

    # Convert ambiguities back to radians and add them to the wrapped phase.
    return wrapped + 2 * np.pi * ambiguity


def unwrap(
    ifg_filename: Filename,
    corr_filename: Filename,
    unw_filename: Filename,
    nlooks: float,
    mask_filename: Optional[Filename] = None,
    unwrap_options: UnwrapOptions = DEFAULT_OPTIONS,
    log_to_file: bool = True,
    unw_nodata: float | None = DEFAULT_UNW_NODATA,
    ccl_nodata: int | None = DEFAULT_CCL_NODATA,
    scratchdir: Optional[Filename] = None,
    delete_scratch: bool = False,
) -> tuple[Path, Path]:
    """Unwrap a single interferogram.

    Parameters
    ----------
    ifg_filename : Filename
        Path to input interferogram.
    corr_filename : Filename
        Path to input correlation file.
    unw_filename : Filename
        Path to output unwrapped phase file.
    unwrap_options : UnwrapOptions, optional
        [`UnwrapOptions`][dolphin.workflows.config.UnwrapOptions] config object
        with parameters and settings for unwrapping.
    nlooks : float
        Effective number of looks used to form the input correlation data.
    mask_filename : Filename, optional
        Path to binary byte mask file, by default None.
        Assumes that 1s are valid pixels and 0s are invalid.
    log_to_file : bool, optional
        Redirect isce3 logging output to file, by default True
    unw_nodata : float , optional.
        Requested nodata value for the unwrapped phase.
        Default = 0
    ccl_nodata : float, optional
        Requested nodata value for connected component labels.
        Default = max value of UInt16 (65535)
    scratchdir : Filename, optional
        Path to scratch directory to hold intermediate files.
        If None, uses `tophu`'s `/tmp/...` default.
    delete_scratch : bool, default = False
        After unwrapping, delete the contents inside `scratchdir`.

    Returns
    -------
    unw_path : Path
        Path to output unwrapped phase file.
    conncomp_path : Path
        Path to output connected component label file.

    """
    unwrap_method = unwrap_options.unwrap_method
    preproc_options = unwrap_options.preprocess_options
    if scratchdir is None:
        # Let the unwrappers handle the scratch if we don't specify.
        delete_scratch = False
    else:
        Path(scratchdir).mkdir(parents=True, exist_ok=True)

    # Check for a nodata mask
    if io.get_raster_nodata(ifg_filename) is None or mask_filename is None:
        # With no marked `nodata`, just use the passed in mask
        combined_mask_file = mask_filename
    else:
        combined_mask_file = Path(ifg_filename).with_suffix(".mask.tif")
        create_combined_mask(
            mask_filename=mask_filename,
            image_filename=ifg_filename,
            output_filename=combined_mask_file,
        )

    unwrapper_ifg_filename = Path(ifg_filename)
    unwrapper_unw_filename = Path(unw_filename)
    name_change = "."

    ifg = io.load_gdal(ifg_filename, masked=True)
    if unwrap_options.run_goldstein:
        suf = Path(unw_filename).suffix
        if suf == ".tif":
            driver = "GTiff"
            opts = list(io.DEFAULT_TIFF_OPTIONS)
        else:
            driver = "ENVI"
            opts = list(io.DEFAULT_ENVI_OPTIONS)

        name_change = ".filt" + name_change
        # If we're running Goldstein filtering, the intermediate
        # filtered/unwrapped rasters are temporary rasters in the scratch dir.
        filt_ifg_filename = Path(scratchdir or ".") / (
            Path(ifg_filename).stem.split(".")[0] + (name_change + "int" + suf)
        )
        filt_unw_filename = Path(
            str(unw_filename).split(".")[0] + (name_change + "unw" + suf)
        )

        logger.info(f"Goldstein filtering {ifg_filename} -> {filt_ifg_filename}")
        modified_ifg = goldstein(ifg.filled(0), alpha=preproc_options.alpha)
        logger.info(f"Writing filtered output to {filt_ifg_filename}")
        io.write_arr(
            arr=modified_ifg,
            output_name=filt_ifg_filename,
            like_filename=ifg_filename,
            driver=driver,
            options=opts,
        )
        unwrapper_ifg_filename = filt_ifg_filename
        unwrapper_unw_filename = filt_unw_filename

    if unwrap_options.run_interpolation:
        suf = Path(ifg_filename).suffix
        if suf == ".tif":
            driver = "GTiff"
            opts = list(io.DEFAULT_TIFF_OPTIONS)
        else:
            driver = "ENVI"
            opts = list(io.DEFAULT_ENVI_OPTIONS)

        pre_interp_ifg_filename = unwrapper_ifg_filename
        pre_interp_unw_filename = unwrapper_unw_filename
        name_change = ".interp" + name_change

        # temporarily storing the intermediate interpolated rasters in the scratch dir.
        interp_ifg_filename = Path(scratchdir or ".") / (
            pre_interp_ifg_filename.stem.split(".")[0] + (name_change + "int" + suf)
        )
        interp_unw_filename = Path(
            str(pre_interp_unw_filename).split(".")[0] + (name_change + "unw" + suf)
        )

        pre_interp_ifg = io.load_gdal(pre_interp_ifg_filename)
        corr = io.load_gdal(corr_filename)
        cutoff = preproc_options.interpolation_cor_threshold
        logger.info(f"Masking pixels with correlation below {cutoff}")
        coherent_pixel_mask = corr[:] >= cutoff

        logger.info(f"Interpolating {pre_interp_ifg_filename} -> {interp_ifg_filename}")
        modified_ifg = interpolate(
            ifg=pre_interp_ifg,
            weights=coherent_pixel_mask,
            weight_cutoff=cutoff,
            max_radius=preproc_options.max_radius,
        )

        logger.info(f"Writing interpolated output to {interp_ifg_filename}")
        io.write_arr(
            arr=modified_ifg,
            output_name=interp_ifg_filename,
            like_filename=ifg_filename,
            driver=driver,
            options=opts,
        )
        unwrapper_ifg_filename = interp_ifg_filename
        unwrapper_unw_filename = interp_unw_filename

    if unwrap_method == UnwrapMethod.SNAPHU:
        snaphu_opts = unwrap_options.snaphu_options
        # Pass everything to snaphu-py
        unw_path, conncomp_path = unwrap_snaphu_py(
            unwrapper_ifg_filename,
            corr_filename,
            unwrapper_unw_filename,
            nlooks,
            ntiles=snaphu_opts.ntiles,
            tile_overlap=snaphu_opts.tile_overlap,
            mask_file=combined_mask_file,
            nproc=snaphu_opts.n_parallel_tiles,
            zero_where_masked=unwrap_options.zero_where_masked,
            unw_nodata=unw_nodata,
            ccl_nodata=ccl_nodata,
            init_method=snaphu_opts.init_method,
            cost=snaphu_opts.cost,
            scratchdir=scratchdir,
        )
    elif unwrap_method == UnwrapMethod.WHIRLWIND:
        unw_path, conncomp_path = unwrap_whirlwind(
            unwrapper_ifg_filename,
            corr_filename,
            unwrapper_unw_filename,
            nlooks,
            mask_file=combined_mask_file,
            zero_where_masked=unwrap_options.zero_where_masked,
            unw_nodata=unw_nodata,
            ccl_nodata=ccl_nodata,
            scratchdir=scratchdir,
        )
    elif (unwrap_method == UnwrapMethod.ICU) or (unwrap_method == UnwrapMethod.PHASS):
        tophu_opts = unwrap_options.tophu_options
        unw_path, conncomp_path = multiscale_unwrap(
            unwrapper_ifg_filename,
            corr_filename,
            unwrapper_unw_filename,
            tophu_opts.downsample_factor,
            ntiles=tophu_opts.ntiles,
            nlooks=nlooks,
            mask_file=combined_mask_file,
            zero_where_masked=unwrap_options.zero_where_masked,
            unw_nodata=unw_nodata,
            ccl_nodata=ccl_nodata,
            init_method=tophu_opts.init_method,
            cost=tophu_opts.cost,
            unwrap_method=unwrap_method,
            scratchdir=scratchdir,
            log_to_file=log_to_file,
        )
    else:
        # Should be unreachable.
        raise AssertionError(f"unexpected unwrap method {unwrap_method}")

    # post-processing steps go here:

    # Transfer ambiguity numbers from filtered/interpolated unwrapped interferogram
    # back to original interferogram
    if unwrap_options.run_goldstein or unwrap_options.run_interpolation:
        logger.info(
            "Transferring ambiguity numbers from filtered/interpolated"
            f" ifg {unwrapper_unw_filename}"
        )
        unw_arr = io.load_gdal(unwrapper_unw_filename, masked=True).filled(unw_nodata)

        final_arr = transfer_ambiguities(np.angle(ifg), unw_arr)
        final_arr[ifg.mask] = unw_nodata

        io.write_arr(
            arr=final_arr,
            output_name=unw_filename,
            like_filename=unwrapper_unw_filename,
            dtype=np.float32,
            driver=driver,
            options=opts,
        )

        # Regrow connected components after phase modification
        corr = io.load_gdal(corr_filename)
        mask = corr[:] > 0
        # TODO decide whether we want to have the
        # 'min_conncomp_frac' option in the config
        conncomp_path = grow_conncomp_snaphu(
            unw_filename=unw_filename,
            corr_filename=corr_filename,
            nlooks=nlooks,
            mask=mask,
            ccl_nodata=ccl_nodata,
            cost=unwrap_options.snaphu_options.cost,
            scratchdir=scratchdir,
        )

        # Move the intermediate ".interp" or ".goldstein" into the scratch directory
        if scratchdir is not None:
            shutil.move(unwrapper_unw_filename, scratchdir)

    # Reset the input nodata values to be nodata in the unwrapped and CCL
    logger.info(f"Setting nodata values of {unw_path} file")
    set_nodata_values(
        filename=unw_filename, output_nodata=unw_nodata, like_filename=ifg_filename
    )
    logger.info(f"Setting nodata values of {conncomp_path} file")
    set_nodata_values(
        filename=conncomp_path, output_nodata=ccl_nodata, like_filename=ifg_filename
    )

    if delete_scratch:
        assert scratchdir is not None
        shutil.rmtree(scratchdir)

    return Path(unw_filename), conncomp_path



from __future__ import annotations

import csv
from functools import lru_cache
from math import log
from pathlib import Path
from typing import Optional

import numba
import numpy as np
from numpy.typing import ArrayLike

from dolphin._types import Strides
from dolphin.utils import _get_slices, compute_out_shape

from ._common import remove_unconnected

_get_slices = numba.njit(_get_slices)


def estimate_neighbors(
    mean: ArrayLike,
    var: ArrayLike,
    halfwin_rowcol: tuple[int, int],
    nslc: int,
    strides: Optional[dict] = None,
    alpha: float = 0.05,
    prune_disconnected: bool = False,
):
    """Estimate the number of neighbors based on the GLRT.

    Based on the method described in [@Parizzi2011AdaptiveInSARStack].
    Assumes Rayleigh distributed amplitudes ([@Siddiqui1962ProblemsConnectedRayleigh])

    Parameters
    ----------
    mean : ArrayLike, 2D
        Mean amplitude of each pixel.
    var: ArrayLike, 2D
        Variance of each pixel's amplitude.
    halfwin_rowcol : tuple[int, int]
        Half the size of the block in (row, col) dimensions
    nslc : int
        Number of images in the stack used to compute `mean` and `var`.
        Used to compute the degrees of freedom for the t- and F-tests to
        determine the critical values.
    strides: dict, optional
        The (x, y) strides (in pixels) to use for the sliding window.
        By default {"x": 1, "y": 1}
    alpha : float, default=0.05
        Significance level at which to reject the null hypothesis.
        Rejecting means declaring a neighbor is not a SHP.
    prune_disconnected : bool, default=False
        If True, keeps only SHPs that are 8-connected to the current pixel.
        Otherwise, any pixel within the window may be considered an SHP, even
        if it is not directly connected.


    Notes
    -----
    When `strides` is not (1, 1), the output first two dimensions
    are smaller than `mean` and `var` by a factor of `strides`. This
    will match the downstream shape of the strided phase linking results.

    Returns
    -------
    is_shp : np.ndarray, 4D
        Boolean array marking which neighbors are SHPs for each pixel in the block.
        Shape is (out_rows, out_cols, window_rows, window_cols), where
            `out_rows` and `out_cols` are computed by
            `[dolphin.io.compute_out_shape][]`
            `window_rows = 2 * halfwin_rowcol[0] + 1`
            `window_cols = 2 * halfwin_rowcol[1] + 1`

    """
    if strides is None:
        strides = {"x": 1, "y": 1}
    half_row, half_col = halfwin_rowcol
    rows, cols = mean.shape

    threshold = get_cutoff(alpha=alpha, N=nslc)

    strides_rowcol = (strides["y"], strides["x"])
    out_rows, out_cols = compute_out_shape((rows, cols), Strides(*strides_rowcol))
    is_shp = np.zeros(
        (out_rows, out_cols, 2 * half_row + 1, 2 * half_col + 1), dtype=np.bool_
    )
    return _loop_over_pixels(
        mean,
        var,
        halfwin_rowcol,
        strides_rowcol,
        threshold,
        prune_disconnected,
        is_shp,
    )


def get_cutoff(alpha: float, N: int) -> float:
    r"""Compute the upper cutoff for the GLRT test statistic.

    Statistic is

    \[
    2\log(\sigma_{pooled}) - \log(\sigma_{p}) -\log(\sigma_{q})
    \]

    Parameters
    ----------
    alpha: float
        Significance level (0 < alpha < 1).
    N: int
        Number of samples.

    Returns
    -------
    float
        Cutoff value for the GLRT test statistic.

    """
    n_alpha_to_cutoff = _read_cutoff_csv()
    try:
        return n_alpha_to_cutoff[(N, alpha)]
    except KeyError as e:
        msg = f"Not implemented for {N = }, {alpha = }"
        raise NotImplementedError(msg) from e


@numba.njit(nogil=True)
def _compute_glrt_test_stat(scale_1, scale_2):
    """Compute the GLRT test statistic."""
    scale_pooled = (scale_1 + scale_2) / 2
    return 2 * log(scale_pooled) - log(scale_1) - log(scale_2)


@numba.njit(nogil=True, parallel=True)
def _loop_over_pixels(
    mean: ArrayLike,
    var: ArrayLike,
    halfwin_rowcol: tuple[int, int],
    strides_rowcol: tuple[int, int],
    threshold: float,
    prune_disconnected: bool,
    is_shp: np.ndarray,
) -> np.ndarray:
    """Loop common to SHP tests using only mean and variance."""
    half_row, half_col = halfwin_rowcol
    row_strides, col_strides = strides_rowcol
    # location to start counting from in the larger input
    r0, c0 = row_strides // 2, col_strides // 2
    in_rows, in_cols = mean.shape
    out_rows, out_cols = is_shp.shape[:2]

    # Convert mean/var to the Rayleigh scale parameter
    scale_squared = (var + mean**2) / 2

    for out_r in numba.prange(out_rows):
        for out_c in range(out_cols):
            in_r = r0 + out_r * row_strides
            in_c = c0 + out_c * col_strides

            scale_1 = scale_squared[in_r, in_c]
            # Clamp the window to the image bounds
            (r_start, r_end), (c_start, c_end) = _get_slices(
                half_row, half_col, in_r, in_c, in_rows, in_cols
            )
            if mean[in_r, in_c] == 0:
                # Skip nodata pixels
                continue

            for in_r2 in range(r_start, r_end):
                for in_c2 in range(c_start, c_end):
                    # window offsets for dims 3,4 of `is_shp`
                    r_off = in_r2 - r_start
                    c_off = in_c2 - c_start

                    # Don't count itself as a neighbor
                    if in_r2 == in_r and in_c2 == in_c:
                        is_shp[out_r, out_c, r_off, c_off] = False
                        continue
                    scale_2 = scale_squared[in_r2, in_c2]

                    T = _compute_glrt_test_stat(scale_1, scale_2)

                    is_shp[out_r, out_c, r_off, c_off] = threshold > T
            if prune_disconnected:
                # For this pixel, prune the groups not connected to the center
                remove_unconnected(is_shp[out_r, out_c], inplace=True)

    return is_shp


@lru_cache
def _read_cutoff_csv():
    filename = Path(__file__).parent / "glrt_cutoffs.csv"

    result = {}
    with open(filename) as file:
        reader = csv.DictReader(file)
        for row in reader:
            n = int(row["N"])
            alpha = float(row["alpha"])
            cutoff = float(row["cutoff"])
            result[(n, alpha)] = cutoff

    return result
